import logging
from dataclasses import dataclass

from mstrio import config
from mstrio.api import applications
from mstrio.connection import Connection
from mstrio.helpers import IServerError
from mstrio.project_objects.palette import Palette
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity
from mstrio.utils.helper import Dictable, delete_none_values, find_object_with_name
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.1200')
def list_applications(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    name: str | None = None,
) -> list['Application'] | list[dict]:
    """Get list of available Applications.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        to_dictionary (bool, optional): if True, return Applications as a
            list of dicts
        limit (int, optional): Limit the number of elements returned
        name (str, optional):  filter for applications with names containing
            this value

    Returns:
        List of Application objects or list of dictionaries containing
        application properties.
    """
    response = (
        applications.get_applications(connection=connection).json().get('applications')
    )
    if name:
        response = [app for app in response if name in app.get('name')]
    if limit:
        response = response[:limit]
    if to_dictionary:
        return response
    else:
        return [
            Application.from_dict(source=obj, connection=connection) for obj in response
        ]


@class_version_handler('11.3.1200')
class Application(Entity, CopyMixin, DeleteMixin):
    """Python representation of a Strategy One Application object"""

    @dataclass
    class GeneralSettings(Dictable):
        """General settings of the application.

        Attributes:
            disable_advanced_settings (bool): if True, hides advanced settings
                menu item
            disable_preferences (bool): if True, hides preferences menu item
            network_timeout (int): network timeout time in seconds
            cache_clear_mode (int): cache clear mode
                Possible values:
                    1: automatic
                    2: on close
            clear_cache_on_logout (bool): if True, clears cache on logout
            max_log_size (int): maximum log size
            log_level (int): log level
                Possible values:
                    0: all
                    10: info
                    12: warning
                    14: severe
                    16: off
            update_interval (int): update interval in minutes
        """

        disable_advanced_settings: bool | None = None
        disable_preferences: bool | None = None
        network_timeout: int | None = None
        cache_clear_mode: int | None = None
        clear_cache_on_logout: bool | None = None
        max_log_size: int | None = None
        log_level: int | None = None
        update_interval: int | None = None

    @dataclass
    class HomeSettings(Dictable):
        """Home Screen settings of the application.

        Attributes:
            mode (str): home screen mode
                Possible values:
                    -0: use Library (default)
                    -1: use Dashboard/Document
                    -2: use Agent
            home_document (HomeDocument): home document settings
            home_library (HomeLibrary): home library settings
            theme (Theme): theme settings
        """

        @dataclass
        class HomeDocument(Dictable):
            """Home Screen as Document settings of the application.

            Attributes:
                url (str): URL of the document
                home_document_type (str): home document type
                    Possible values:
                        -dashboard
                        -document
                icons (list[str]): list of icons to be shown on the home screen
                    toolbar
                    Possible values:
                        -comments
                        -notifications
                        -options
                        -table_of_contents
                        -bookmarks
                        -reset
                        -filters
                        -share
                        -data_search
                        -hyper_intelligence
                        -font_size
                toolbar_mode (str): toolbar mode
                    Possible values:
                        -0: show toolbar (default)
                        -1: toolbar will be collapsed by default and expanded
                            by tapping the top right corner icon
                toolbar_enabled (bool): whether the toolbar is enabled
            """

            url: str | None = None
            home_document_type: str | None = None
            icons: list[str] | None = None
            toolbar_mode: str | None = None
            toolbar_enabled: bool | None = None

        @dataclass
        class HomeLibrary(Dictable):
            """Home Screen as Library settings of the application.

            Attributes:
                content_bundle_ids (list[str]): list of content bundle IDs that
                    the application is limited to
                show_all_contents (bool): whether to show all library contents
                    if the content groups are specified in the application
                icons (list[str]): list of icons to be shown on the home screen
                    toolbar
                    Possible values:
                        -comments
                        -notifications
                        -options
                        -table_of_contents
                        -bookmarks
                        -reset
                        -filters
                        -share
                        -data_search
                        -hyper_intelligence
                        -font_size
                customized_items (CustomizedItems): customized settings
                customized_item_properties (dict): extra properties of the items
                    that were customized
                toolbar_mode (str): toolbar mode
                    Possible values:
                        -0: show toolbar (default)
                        -1: toolbar will be collapsed by default and expanded
                            by tapping the top right corner icon
                sidebars (list[str]): menus to be shown in the library sidebar
                    Possible values:
                        -all
                        -favorites
                        -recents
                        -default_groups
                        -my_groups
                        -options
                toolbar_enabled (bool): whether the toolbar is enabled
                default_groups_name (str): customized name for default groups
            """

            @dataclass
            class CustomizedItems(Dictable):
                """Customized items settings of the application."""

                def __init__(
                    self,
                    my_content: bool | None = None,
                    subscriptions: bool | None = None,
                    new_dashboard: bool | None = None,
                    edit_dashboard: bool | None = None,
                    add_library_server: bool | None = None,
                    data_search: bool | None = None,
                    hyper_intelligence: bool | None = None,
                    font_size: bool | None = None,
                    undo_and_redo: bool | None = None,
                    insights: bool | None = None,
                    content_discovery: bool | None = None,
                    mobile_account_panel_user_name: bool | None = None,
                    mobile_account_panel_preferences_my_language: bool | None = None,
                    mobile_account_panel_preferences_my_time_zone: bool | None = None,
                    mobile_account_panel_preferences_face_id_login: bool | None = None,
                    mobile_account_panel_preferences_take_a_tour: bool | None = None,
                    mobile_account_panel_preferences_refresh_view_automatically: (
                        bool | None
                    ) = None,
                    mobile_account_panel_preferences_smart_download: bool | None = None,
                    mobile_account_panel_preferences_automatically_add_to_library: (
                        bool | None
                    ) = None,
                    mobile_account_panel_advanced_settings_app_settings: (
                        bool | None
                    ) = None,
                    mobile_account_panel_advanced_settings_security_settings: (
                        bool | None
                    ) = None,
                    mobile_account_panel_advanced_settings_logging: bool | None = None,
                    mobile_account_panel_help_and_legal: bool | None = None,
                    mobile_account_panel_help_and_legal_help: bool | None = None,
                    mobile_account_panel_help_and_legal_legal: bool | None = None,
                    mobile_account_panel_help_and_legal_report_a_problem: (
                        bool | None
                    ) = None,
                    mobile_account_panel_log_out: bool | None = None,
                    filter_summary: bool | None = None,
                    share_panel_share: bool | None = None,
                    share_panel_export_to_excel: bool | None = None,
                    share_panel_export_to_pdf: bool | None = None,
                    share_panel_download: bool | None = None,
                    share_panel_subscribe: bool | None = None,
                    share_panel_annotate_and_share: bool | None = None,
                    web_account_panel_user_name: bool | None = None,
                    web_account_panel_my_library: bool | None = None,
                    web_account_panel_manage_library: bool | None = None,
                    web_account_panel_preference: bool | None = None,
                    web_account_panel_preference_my_language: bool | None = None,
                    web_account_panel_preference_my_time_zone: bool | None = None,
                    web_account_panel_switch_workspace: bool | None = None,
                    web_account_panel_take_a_tour: bool | None = None,
                    web_account_panel_help: bool | None = None,
                    web_account_panel_log_out: bool | None = None,
                    mobile_downloads: bool | None = None,
                    table_of_contents_header: bool | None = None,
                    table_of_contents_content_info: bool | None = None,
                    table_of_contents_chapter_and_page: bool | None = None,
                    switch_library_server: bool | None = None,
                    create_new_content_dashboard: bool | None = None,
                    create_new_content_report: bool | None = None,
                    layout_tile_view: bool | None = None,
                    layout_list_view: bool | None = None,
                    ai_assistant: bool | None = None,
                    share_panel_manage_access: bool | None = None,
                    bot_window_share_panel: bool | None = None,
                    bot_window_share_panel_share_bot: bool | None = None,
                    bot_window_share_panel_embed_bot: bool | None = None,
                    bot_window_share_panel_manage_access: bool | None = None,
                    bot_window_edit_bot: bool | None = None,
                    create_new_content_bot: bool | None = None,
                    dashboard_view_mode: bool | None = None,
                    content_info_content_creator: bool | None = None,
                    content_info_timestamp: bool | None = None,
                    content_info_description: bool | None = None,
                    content_info_project: bool | None = None,
                    content_info_path: bool | None = None,
                    content_info_object_id: bool | None = None,
                    content_info_info_window: bool | None = None,
                    control_filter_summary: bool | None = None,
                    hide_filter_summary: bool | None = None,
                    sidebars_unpin: bool | None = None,
                    table_of_contents_unpin: bool | None = None,
                    filter_panel_unpin: bool | None = None,
                    comments_panel_unpin: bool | None = None,
                    ai_assistant_unpin: bool | None = None,
                    table_of_contents_allow_close: bool | None = None,
                    filter_panel_allow_close: bool | None = None,
                    comments_panel_allow_close: bool | None = None,
                    ai_assistant_allow_close: bool | None = None,
                    use_application_name_as_library_title: bool | None = None,
                    **kwargs,
                ):
                    self.my_content = my_content
                    self.subscriptions = subscriptions
                    self.new_dossier = new_dashboard
                    self.edit_dossier = edit_dashboard
                    self.add_library_server = add_library_server
                    self.data_search = data_search
                    self.hyper_intelligence = hyper_intelligence
                    self.font_size = font_size
                    self.undo_and_redo = undo_and_redo
                    self.insights = insights
                    self.content_discovery = content_discovery
                    self.mobile_account_panel_user_name = mobile_account_panel_user_name
                    self.mobile_account_panel_preferences_my_language = (
                        mobile_account_panel_preferences_my_language
                    )
                    self.mobile_account_panel_preferences_my_time_zone = (
                        mobile_account_panel_preferences_my_time_zone
                    )
                    self.mobile_account_panel_preferences_face_id_login = (
                        mobile_account_panel_preferences_face_id_login
                    )
                    self.mobile_account_panel_preferences_take_a_tour = (
                        mobile_account_panel_preferences_take_a_tour
                    )
                    self.mobile_account_panel_preferences_refresh_view_automatically = (
                        mobile_account_panel_preferences_refresh_view_automatically
                    )
                    self.mobile_account_panel_preferences_smart_download = (
                        mobile_account_panel_preferences_smart_download
                    )
                    self.mobile_account_panel_preferences_automatically_add_to_library = (  # noqa: E501
                        mobile_account_panel_preferences_automatically_add_to_library
                    )
                    self.mobile_account_panel_advanced_settings_app_settings = (
                        mobile_account_panel_advanced_settings_app_settings
                    )
                    self.mobile_account_panel_advanced_settings_security_settings = (
                        mobile_account_panel_advanced_settings_security_settings
                    )
                    self.mobile_account_panel_advanced_settings_logging = (
                        mobile_account_panel_advanced_settings_logging
                    )
                    self.mobile_account_panel_help_and_legal = (
                        mobile_account_panel_help_and_legal
                    )
                    self.mobile_account_panel_help_and_legal_help = (
                        mobile_account_panel_help_and_legal_help
                    )
                    self.mobile_account_panel_help_and_legal_legal = (
                        mobile_account_panel_help_and_legal_legal
                    )
                    self.mobile_account_panel_help_and_legal_report_a_problem = (
                        mobile_account_panel_help_and_legal_report_a_problem
                    )
                    self.mobile_account_panel_log_out = mobile_account_panel_log_out
                    self.filter_summary = filter_summary
                    self.share_panel_share = share_panel_share
                    self.share_panel_export_to_excel = share_panel_export_to_excel
                    self.share_panel_export_to_pdf = share_panel_export_to_pdf
                    self.share_panel_download = share_panel_download
                    self.share_panel_subscribe = share_panel_subscribe
                    self.share_panel_annotate_and_share = share_panel_annotate_and_share
                    self.web_account_panel_user_name = web_account_panel_user_name
                    self.web_account_panel_my_library = web_account_panel_my_library
                    self.web_account_panel_manage_library = (
                        web_account_panel_manage_library
                    )
                    self.web_account_panel_preference = web_account_panel_preference
                    self.web_account_panel_preference_my_language = (
                        web_account_panel_preference_my_language
                    )
                    self.web_account_panel_preference_my_time_zone = (
                        web_account_panel_preference_my_time_zone
                    )
                    self.web_account_panel_switch_workspace = (
                        web_account_panel_switch_workspace
                    )
                    self.web_account_panel_take_a_tour = web_account_panel_take_a_tour
                    self.web_account_panel_help = web_account_panel_help
                    self.web_account_panel_log_out = web_account_panel_log_out
                    self.mobile_downloads = mobile_downloads
                    self.table_of_contents_header = table_of_contents_header
                    self.table_of_contents_content_info = table_of_contents_content_info
                    self.table_of_contents_chapter_and_page = (
                        table_of_contents_chapter_and_page
                    )
                    self.switch_library_server = switch_library_server
                    self.create_new_content_dossier = create_new_content_dashboard
                    self.create_new_content_report = create_new_content_report
                    self.layout_tile_view = layout_tile_view
                    self.layout_list_view = layout_list_view
                    self.ai_assistant = ai_assistant
                    self.share_panel_manage_access = share_panel_manage_access
                    self.bot_window_share_panel = bot_window_share_panel
                    self.bot_window_share_panel_share_bot = (
                        bot_window_share_panel_share_bot
                    )
                    self.bot_window_share_panel_embed_bot = (
                        bot_window_share_panel_embed_bot
                    )
                    self.bot_window_share_panel_manage_access = (
                        bot_window_share_panel_manage_access
                    )
                    self.bot_window_edit_bot = bot_window_edit_bot
                    self.create_new_content_bot = create_new_content_bot
                    self.dashboard_view_mode = dashboard_view_mode
                    self.content_info_content_creator = content_info_content_creator
                    self.content_info_timestamp = content_info_timestamp
                    self.content_info_description = content_info_description
                    self.content_info_project = content_info_project
                    self.content_info_path = content_info_path
                    self.content_info_object_id = content_info_object_id
                    self.content_info_info_window = content_info_info_window
                    self.control_filter_summary = control_filter_summary
                    self.hide_filter_summary = hide_filter_summary
                    self.sidebars_unpin = sidebars_unpin
                    self.table_of_contents_unpin = table_of_contents_unpin
                    self.filter_panel_unpin = filter_panel_unpin
                    self.comments_panel_unpin = comments_panel_unpin
                    self.ai_assistant_unpin = ai_assistant_unpin
                    self.table_of_contents_allow_close = table_of_contents_allow_close
                    self.filter_panel_allow_close = filter_panel_allow_close
                    self.comments_panel_allow_close = comments_panel_allow_close
                    self.ai_assistant_allow_close = ai_assistant_allow_close
                    self.use_application_name_as_library_title = (
                        use_application_name_as_library_title
                    )

                    possible_new_args = [
                        'new_dossier',
                        'edit_dossier',
                        'create_new_content_dossier',
                    ]
                    if kwargs and not all(
                        key in possible_new_args for key in kwargs.keys()
                    ):
                        raise KeyError(
                            f"Invalid kwargs keys: {list(kwargs.keys())} "
                            f"provided to '{self.__class__.__name__}' class. "
                            f"Possible kwargs keys: {possible_new_args}"
                        )

                    if new_dossier := kwargs.get('new_dossier'):
                        Application._check_replaced_properties(
                            new_dashboard, new_dossier, 'new_dashboard', 'new_dossier'
                        )
                        self.new_dossier = new_dossier

                    if edit_dossier := kwargs.get('edit_dossier'):
                        Application._check_replaced_properties(
                            edit_dashboard,
                            edit_dossier,
                            'edit_dashboard',
                            'edit_dossier',
                        )
                        self.edit_dossier = edit_dossier

                    if create_new_content_dossier := kwargs.get(
                        'create_new_content_dossier'
                    ):
                        Application._check_replaced_properties(
                            create_new_content_dashboard,
                            create_new_content_dossier,
                            'create_new_content_dashboard',
                            'create_new_content_dossier',
                        )
                        self.create_new_content_dossier = create_new_content_dossier

            _FROM_DICT_MAP = {
                'customized_items': CustomizedItems.from_dict,
            }

            content_bundle_ids: list[str] | None = None
            show_all_contents: bool | None = None
            icons: list[str] | None = None
            customized_items: CustomizedItems | None = None
            customized_item_properties: dict | None = None
            toolbar_mode: str | None = None
            sidebars: list[str] | None = None
            toolbar_enabled: bool | None = None
            default_groups_name: str | None = None

        @dataclass
        class Theme(Dictable):
            """Theme settings of the application.

            Attributes:
                logos (Logos): logos settings
                color (Color): color settings
            """

            @dataclass
            class Logos(Dictable):
                """Logos settings of the application.

                Attributes:
                    web (Logo): web logo settings
                    favicon (Logo): favorite icon settings
                    mobile (Logo): mobile logo settings
                """

                @dataclass
                class Logo(Dictable):
                    """Details of a logo.

                    Attributes:
                        type (str): type of the logo
                        value (str): value of the logo
                    """

                    type: str | None = None
                    value: str | None = None

                _FROM_DICT_MAP = {
                    'web': Logo.from_dict,
                    'favicon': Logo.from_dict,
                    'mobile': Logo.from_dict,
                }

                web: Logo | None = None
                favicon: Logo | None = None
                mobile: Logo | None = None

            @dataclass
            class Color(Dictable):
                """Color settings of the application.

                Attributes:
                    selected_theme (str): selected theme
                        Possible values:
                            -useSystemSettings
                            -light
                            -dark
                            -red
                            -green
                            -blue
                            -darkBlue
                            -yellow
                            -custom
                    formatting (Formatting): formatting settings
                    enable_for_bots (bool): whether the theme is enabled for
                        bots
                """

                @dataclass
                class Formatting(Dictable):
                    """Formatting settings of the application.
                    All color values are in hex format.

                    Attributes:
                        toolbar_fill (str): toolbar background color
                        toolbar_color (str): toolbar icon/text color
                        sidebar_fill (str): sidebar background color
                        sidebar_color (str): sidebar text color
                        sidebar_active_fill (str): active sidebar background
                            color
                        sidebar_active_color (str): active sidebar text color
                        panel_fill (str): panel background color
                        panel_color (str): panel text color
                        accent_fill (str): accent fill color
                        notification_badge_fill (str): notification badge color
                        button_color (str): button color
                        canvas_fill (str): canvas background color
                    """

                    toolbar_fill: str | None = None
                    toolbar_color: str | None = None
                    sidebar_fill: str | None = None
                    sidebar_color: str | None = None
                    sidebar_active_fill: str | None = None
                    sidebar_active_color: str | None = None
                    panel_fill: str | None = None
                    panel_color: str | None = None
                    accent_fill: str | None = None
                    notification_badge_fill: str | None = None
                    button_color: str | None = None
                    canvas_fill: str | None = None

                _FROM_DICT_MAP = {
                    'formatting': Formatting.from_dict,
                }

                selected_theme: str | None = None
                formatting: Formatting | None = None
                enable_for_bots: bool | None = None

            _FROM_DICT_MAP = {
                'logos': Logos.from_dict,
                'color': Color.from_dict,
            }

            logos: Logos | None = None
            color: Color | None = None

        _FROM_DICT_MAP = {
            'home_document': HomeDocument.from_dict,
            'home_library': HomeLibrary.from_dict,
            'theme': Theme.from_dict,
        }

        mode: str | None = None
        home_document: HomeDocument | None = None
        home_library: HomeLibrary | None = None
        theme: Theme | None = None

    @dataclass
    class EmailSettings(Dictable):
        """Email settings of the application.

        Attributes:
            enabled (bool): whether to enable custom email settings
            host_portal (str): the host web portal in the button link
            show_branding_image (bool): if True, shows branding image
            show_browser_button (bool): if True, shows browser button
            show_mobile_button (bool): if True, shows mobile button
            show_button_description (bool): if True, shows the email button
                description
            show_reminder (bool): if True, shows reminder section
            show_sent_by (bool): if True, shows sent by section
            sent_by_text (str): text in the sent by section
            show_social_media (bool): if True, shows social media section
            content (Content): content settings
            sender (Sender): sender settings
            branding_image (str): url of the branding image
            button (Button): button settings
            reminder (Reminder): reminder settings
            social_media (SocialMedia): social media settings
        """

        class Content(Dictable):
            """Content settings of the application.

            Attributes:
                share_dashboard (EmailDetails): settings for sharing a dashboard
                share_bookmark (EmailDetails): settings for sharing a dossier
                    with bookmark
                share_bot (EmailDetails): settings for sharing a bot
                member_added (EmailDetails): settings for inviting a recipient
                    to a discussion
                user_mention (EmailDetails): settings for mentioning a recipient
                    in a comment or discussion
            """

            @dataclass
            class EmailDetails(Dictable):
                """Email details settings of the application.

                Attributes:
                    subject (str): email subject
                    body (str): email body
                """

                subject: str | None = None
                body: str | None = None

            _FROM_DICT_MAP = {
                'share_dossier': EmailDetails.from_dict,
                'share_bookmark': EmailDetails.from_dict,
                'share_bot': EmailDetails.from_dict,
                'member_added': EmailDetails.from_dict,
                'user_mention': EmailDetails.from_dict,
            }

            def __init__(
                self,
                share_dashboard: EmailDetails | None = None,
                share_bookmark: EmailDetails | None = None,
                share_bot: EmailDetails | None = None,
                member_added: EmailDetails | None = None,
                user_mention: EmailDetails | None = None,
                **kwargs,
            ):
                self.share_dossier = share_dashboard
                self.share_bookmark = share_bookmark
                self.share_bot = share_bot
                self.member_added = member_added
                self.user_mention = user_mention

                if kwargs and not all(
                    key in ['share_dossier'] for key in kwargs.keys()
                ):
                    raise KeyError(
                        f"Invalid kwargs keys: {list(kwargs.keys())} provided "
                        f"to '{self.__class__.__name__}' class. Possible "
                        f"kwargs keys: ['share_dossier']"
                    )
                if share_dossier := kwargs.get('share_dossier'):
                    Application._check_replaced_properties(
                        share_dashboard,
                        share_dossier,
                        'share_dashboard',
                        'share_dossier',
                    )
                    self.share_dossier = share_dossier

        @dataclass
        class Sender(Dictable):
            """Sender settings of the application.

            Attributes:
                address (str): sender email address
                display_name (str): sender display name
            """

            address: str | None = None
            display_name: str | None = None

        @dataclass
        class Button(Dictable):
            """Button settings of the application.

            Attributes:
                browser_button_style (ButtonStyle): browser button style
                mobile_button_style (ButtonStyle): mobile button style
                mobile_button_link_type (str): mobile button link type
                    Possible values:
                        -default
                        -app_scheme
                        -universal_link
                mobile_button_scheme (str): mobile button scheme
                description (str): button description
            """

            @dataclass
            class ButtonStyle(Dictable):
                """Button style settings of the application.

                Attributes:
                    background_color (str): button background color
                    font_color (str): button text color
                    text (str): button border color
                """

                background_color: str | None = None
                font_color: str | None = None
                text: str | None = None

            _FROM_DICT_MAP = {
                'browser_button_style': ButtonStyle.from_dict,
                'mobile_button_style': ButtonStyle.from_dict,
            }

            browser_button_style: ButtonStyle | None = None
            mobile_button_style: ButtonStyle | None = None
            mobile_button_link_type: str | None = None
            mobile_button_scheme: str | None = None
            description: str | None = None

        @dataclass
        class Reminder(Dictable):
            """Reminder settings of the application.

            Attributes:
                text (str): reminder text
                link_text (str): reminder link text
            """

            text: str | None = None
            link_text: str | None = None

        @dataclass
        class SocialMedia(Dictable):
            """Social media settings of the application.

            Attributes:
                show_facebook (bool): whether to show the Facebook link
                facebook_link (str): Facebook link
                show_twitter (bool): whether to show the Twitter link
                twitter_link (str): Twitter link
                show_linked_in (bool): whether to show the LinkedIn link
                linked_in_link (str): LinkedIn link
                show_youtube (bool): whether to show the YouTube link
                youtube_link (str): YouTube link
            """

            show_facebook: bool | None = None
            facebook_link: str | None = None
            show_twitter: bool | None = None
            twitter_link: str | None = None
            show_linked_in: bool | None = None
            linked_in_link: str | None = None
            show_you_tube: bool | None = None
            you_tube_link: str | None = None

        _FROM_DICT_MAP = {
            'content': Content.from_dict,
            'sender': Sender.from_dict,
            'button': Button.from_dict,
            'reminder': Reminder.from_dict,
            'social_media': SocialMedia.from_dict,
        }

        enabled: bool | None = None
        host_portal: str | None = None
        show_branding_image: bool | None = None
        show_browser_button: bool | None = None
        show_mobile_button: bool | None = None
        show_button_description: bool | None = None
        show_reminder: bool | None = None
        show_sent_by: bool | None = None
        sent_by_text: str | None = None
        show_social_media: bool | None = None
        content: Content | None = None
        sender: Sender | None = None
        branding_image: dict[str, str] | None = None
        button: Button | None = None
        reminder: Reminder | None = None
        social_media: SocialMedia | None = None

    @dataclass
    class AiSettings(Dictable):
        """AI settings of the application.

        Attributes:
            feedback (bool): if True, AI assistance feedback is enabled
            learning (bool): if True, AI assistance learning is enabled
            disclaimer (str): AI assistance disclaimer
        """

        feedback: bool | None = None
        learning: bool | None = None
        disclaimer: str | None = None

    @dataclass
    class AuthModes(Dictable):
        """Authentication modes of the application.

        Attributes:
            available_modes (list[int]): available authentication modes
                Possible values:
                    1: Standard
                    16: LDAP
                    1048576: SAML
                    4194304: OIDC
            default_mode (int): default authentication mode
        """

        available_modes: list[int] | None = None
        default_mode: int | None = None

    @dataclass
    class Environments(Dictable):
        """Environments of the application.

        Attributes:
            current (str): name of the current library environment
            other (list[EnvironmentLink]): list of other environments
        """

        @dataclass
        class EnvironmentLink(Dictable):
            """Environment link settings of the application.

            Attributes:
                name (str): name of the environment
                url (str): URL of the environment
            """

            name: str | None = None
            url: str | None = None

        _FROM_DICT_MAP = {
            'other': [EnvironmentLink.from_dict],
        }

        current: str | None = None
        other: list[EnvironmentLink] | None = None

    _OBJECT_TYPE = ObjectTypes.APPLICATION
    _API_DELETE = staticmethod(applications.delete_application)
    _API_GETTERS = {
        **Entity._API_GETTERS,
        (
            'managed',
            'property_groups',
            'access_granted',
            'general',
            'home_screen',
            'platforms',
            'application_palettes',
            'application_default_palette',
            'show_builtin_palettes',
            'is_default',
            'use_config_palettes',
            'email_settings',
            'ai_settings',
            'auth_modes',
            'environments',
            'application_nuggets',
            'object_names',
            'object_acl',
            'access_granted',
        ): applications.get_application,
    }

    def __init__(
        self, connection: Connection, name: str | None = None, id: str | None = None
    ) -> None:
        """Initialize Application object by passing name or id.

        Args:
            connection (object): Strategy One connection object returned
                by `connection.Connection()`
            name (string, optional): name of Application
            id (string, optional): ID of Application
        """
        if not id:
            if not name:
                raise ValueError("Please specify either 'name' or 'id'.")
            app = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_applications,
            )
            id = app['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self.managed = kwargs.get('managed')
        self.property_groups = kwargs.get('property_groups')
        self.access_granted = kwargs.get('access_granted')
        self._general = kwargs.get('general')
        self._home_screen = kwargs.get('home_screen')
        self.platforms = kwargs.get('platforms')
        self.application_palettes = kwargs.get('application_palettes')
        self.application_default_palette = kwargs.get('application_default_palette')
        self.show_builtin_palettes = kwargs.get('show_builtin_palettes')
        self.is_default = kwargs.get('is_default')
        self.use_config_palettes = kwargs.get('use_config_palettes')
        self._email_settings = kwargs.get('email_settings')
        self._ai_settings = kwargs.get('ai_settings')
        self._auth_modes = kwargs.get('auth_modes')
        self._environments = kwargs.get('environments')
        self.application_nuggets = kwargs.get('application_nuggets')
        self.object_names = kwargs.get('object_names')
        self.object_acl = kwargs.get('object_acl')
        self.access_granted = kwargs.get('access_granted')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        home_screen: 'Application.HomeSettings',
        description: str | None = None,
        managed: bool | None = None,
        general: 'Application.GeneralSettings | None' = None,
        platforms: list[str] | None = None,
        application_palettes: list[str] | list[Palette] | None = None,
        application_default_palette: str | Palette | None = None,
        show_builtin_palettes: bool | None = None,
        is_default: bool | None = None,
        use_config_palettes: bool | None = None,
        email_settings: 'Application.EmailSettings | None' = None,
        ai_settings: 'Application.AiSettings | None' = None,
        auth_modes: 'Application.AuthModes | None' = None,
        environments: 'Application.Environments | None' = None,
        application_nuggets: list[str] | None = None,
    ) -> 'Application':
        """Create a new application.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            name (str): name of the application
            home_screen (Application.HomeSettings): home screen settings of the
                application
            description (str, optional): description of the application
            managed (bool, optional): whether the application is managed
            general (Application.GeneralSettings, optional): general settings
                of the application
            platforms (list[str], optional): list of platforms for the
                application
                Available values:
                    -web
                    -mobile
                    -desktop
            application_palettes (list[str] or list[Palette], optional): list
                of customized application palettes
            application_default_palette (str or Palette, optional): default
                application palette
            show_builtin_palettes (bool, optional): whether to show built-in
                palettes
            is_default (bool, optional): whether the application configuration
                is default
            use_config_palettes (bool, optional): whether to use default
                configuration of palettes
            email_settings (Application.EmailSettings, optional): email settings
                of the application
            ai_settings (Application.AiSettings, optional): AI settings of the
                application
            auth_modes (Application.AuthModes, optional): authentication modes
                of the application
            environments (Application.Environments, optional): environments of
                the application
            application_nuggets (list[str], optional): list of application
                nuggets

        Returns:
            The created Application object.
        """
        if isinstance(application_default_palette, Palette):
            application_default_palette = application_default_palette.id
        if application_palettes:
            application_palettes = [
                palette.id if isinstance(palette, Palette) else palette
                for palette in application_palettes
            ]
        body = {
            'name': name,
            'description': description,
            'managed': managed,
            'general': general.to_dict() if general else None,
            'homeScreen': home_screen.to_dict(),
            'platforms': platforms,
            'applicationPalettes': application_palettes,
            'applicationDefaultPalette': application_default_palette,
            'showBuiltinPalettes': show_builtin_palettes,
            'isDefault': is_default,
            'useConfigPalettes': use_config_palettes,
            'emailSettings': email_settings.to_dict() if email_settings else None,
            'aiSettings': ai_settings.to_dict() if ai_settings else None,
            'authModes': auth_modes.to_dict() if auth_modes else None,
            'environments': environments.to_dict() if environments else None,
            'applicationNuggets': application_nuggets,
            'objectNames': [],
            'objectAcl': [],
        }
        body['homeScreen']['homeLibrary']['customizedItems'] = (
            (home_screen.home_library.customized_items.to_dict(camel_case=False))
            if home_screen.home_library.customized_items
            else None
        )
        body['homeScreen']['homeLibrary']['customizedItemProperties'] = (
            home_screen.home_library.customized_item_properties
            if home_screen.home_library.customized_item_properties
            else None
        )
        if home_screen.home_document.home_document_type == 'dashboard':
            body['homeScreen']['homeDocument']['homeDocumentType'] = 'dossier'
        if email_settings and email_settings.content:
            body['emailSettings']['content'] = {
                'SHARE_DOSSIER': body['emailSettings']['content']['shareDossier'],
                'SHARE_BOOKMARK': body['emailSettings']['content']['shareBookmark'],
                'SHARE_BOT': body['emailSettings']['content']['shareBot'],
                'MEMBER_ADDED': body['emailSettings']['content']['memberAdded'],
                'USER_MENTION': body['emailSettings']['content']['userMention'],
            }
        body = delete_none_values(
            source=body,
            recursion=True,
            whitelist_attributes=['objectNames', 'objectAcl'],
        )
        try:
            applications.create_application(connection=connection, body=body)
        except IServerError as err:
            if "type 'Palette' does not exist in the metadata" in str(err):
                raise AttributeError(
                    "Application cannot be created: at least one palette parameter "
                    "is either invalid or refers to project-level palette. "
                    "Please select a valid configuration-level palette(s) instead."
                ) from err
            raise err

        if config.verbose:
            logger.info(f'Application "{name}" has been created.')
        # The endpoint returns nothing, so we need to return app manually
        return Application(connection=connection, name=name)

    def alter(
        self,
        home_screen: 'Application.HomeSettings | None' = None,
        name: str | None = None,
        description: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        managed: bool | None = None,
        general: 'Application.GeneralSettings | None' = None,
        platforms: list[str] | None = None,
        application_palettes: list[str] | list[Palette] | None = None,
        application_default_palette: str | Palette | None = None,
        show_builtin_palettes: bool | None = None,
        is_default: bool | None = None,
        use_config_palettes: bool | None = None,
        email_settings: 'Application.EmailSettings | None' = None,
        ai_settings: 'Application.AiSettings | None' = None,
        auth_modes: 'Application.AuthModes | None' = None,
        environments: 'Application.Environments | None' = None,
        application_nuggets: list[str] | None = None,
    ) -> None:
        """Alter an application.

        Args:
            home_screen (Application.HomeSettings, optional): home screen
                settings of the application
            name (str, optional): name of the application
            description (str, optional): description of the application
            comments (str, optional): long description of the application
            owner: (str, User, optional): owner of the application object
            managed (bool, optional): whether the application is managed
            general (Application.GeneralSettings, optional): general settings
                of the application
            platforms (list[str], optional): list of platforms for the
                application
                Available values:
                    -web
                    -mobile
                    -desktop
            application_palettes (list[str] or list[Palette], optional): list
                of customized application palettes
            application_default_palette (str or Palette, optional): default
                application palette
            show_builtin_palettes (bool, optional): whether to show built-in
                palettes
            is_default (bool, optional): whether the application configuration
                is default
            use_config_palettes (bool, optional): whether to use default
                configuration of palettes
            email_settings (Application.EmailSettings, optional): email settings
                of the application
            ai_settings (Application.AiSettings, optional): AI settings of the
                application
            auth_modes (Application.AuthModes, optional): authentication modes
                of the application
            environments (Application.Environments, optional): environments of
                the application
            application_nuggets (list[str], optional): list of application
                nuggets
        """
        if not home_screen:
            home_screen = self.home_screen
        if isinstance(application_default_palette, Palette):
            application_default_palette = application_default_palette.id
        if application_palettes:
            application_palettes = [
                palette.id if isinstance(palette, Palette) else palette
                for palette in application_palettes
            ]
        if isinstance(owner, User):
            owner = owner.id
        body = {
            'objectVersion': self.version,
            'name': name or self.name,
            'description': description or self.description,
            'comments': comments or self.comments,
            'ownerId': owner or self.owner['id'],
            'managed': managed or self.managed,
            'general': (
                (general or self.general).to_dict()
                if (general or self.general)
                else None
            ),
            'homeScreen': home_screen.to_dict(),
            'platforms': platforms or self.platforms,
            'applicationPalettes': application_palettes
            or self.application_palettes
            or [],
            'applicationDefaultPalette': application_default_palette
            or self.application_default_palette
            or '',
            'showBuiltinPalettes': show_builtin_palettes or self.show_builtin_palettes,
            'isDefault': is_default or self.is_default,
            'useConfigPalettes': use_config_palettes or self.use_config_palettes,
            'emailSettings': (
                (email_settings or self.email_settings).to_dict()
                if (email_settings or self.email_settings)
                else None
            ),
            'aiSettings': (
                (ai_settings or self.ai_settings).to_dict()
                if (ai_settings or self.ai_settings)
                else None
            ),
            'authModes': (
                (auth_modes or self.auth_modes).to_dict()
                if (auth_modes or self.auth_modes)
                else None
            ),
            'environments': (
                (environments or self.environments).to_dict()
                if (environments or self.environments)
                else None
            ),
            'applicationNuggets': application_nuggets or self.application_nuggets or [],
            'accessGranted': self.access_granted,
            'objectNames': self.object_names or [],
            'objectAcl': self.object_acl or [],
        }
        if home_screen:
            body['homeScreen']['homeLibrary']['customizedItems'] = (
                (home_screen.home_library.customized_items.to_dict(camel_case=False))
                if home_screen.home_library.customized_items
                else None
            )
            body['homeScreen']['homeLibrary']['customizedItemProperties'] = (
                home_screen.home_library.customized_item_properties or None
            )
            if home_screen.home_document.home_document_type == 'dashboard':
                body['homeScreen']['homeDocument']['homeDocumentType'] = 'dossier'
        if body['emailSettings'] is not None and body['emailSettings'].get('content'):
            body['emailSettings']['content'] = {
                'SHARE_DOSSIER': body['emailSettings']['content']['shareDossier'],
                'SHARE_BOOKMARK': body['emailSettings']['content']['shareBookmark'],
                'SHARE_BOT': body['emailSettings']['content']['shareBot'],
                'MEMBER_ADDED': body['emailSettings']['content']['memberAdded'],
                'USER_MENTION': body['emailSettings']['content']['userMention'],
            }
        body = delete_none_values(
            source=body,
            recursion=True,
            whitelist_attributes=[
                'objectNames',
                'objectAcl',
                'applicationNuggets',
                'applicationPalettes',
            ],
        )
        applications.update_application(
            connection=self.connection, id=self.id, body=body
        )
        if config.verbose:
            logger.info(
                'Application has been altered successfully. Your changes are not '
                'saved locally. Refresh the object to see the changes.'
            )

    @staticmethod
    def _check_replaced_properties(
        dashboard_property, dossier_property, dashboard_name, dossier_name
    ):
        if dashboard_property is not None and dashboard_property != dossier_property:
            raise KeyError(
                f"You provided both '{dashboard_name}' and '{dossier_name}' "
                f"with different values. Please use only {dashboard_name}."
            )

    @property
    def general(self) -> GeneralSettings | None:
        """General settings of the application."""
        if not self._general:
            self.fetch('general')
        if self._general:
            return Application.GeneralSettings.from_dict(source=self._general)
        return None

    @property
    def home_screen(self) -> HomeSettings | None:
        """Home Screen settings of the application."""
        if not self._home_screen:
            self.fetch('home_screen')
        if self._home_screen:
            return Application.HomeSettings.from_dict(source=self._home_screen)
        return None

    @property
    def email_settings(self) -> EmailSettings | None:
        """Email settings of the application."""
        if not self._email_settings:
            self.fetch('email_settings')
        if self._email_settings:
            temp = Application.EmailSettings.from_dict(source=self._email_settings)
            content = self._email_settings.get('content')
            # Initialize Content class only if content is not empty.
            # Otherwise, keep it None.
            if content:
                temp.content = Application.EmailSettings.Content(
                    share_dashboard=content.get('SHARE_DOSSIER'),
                    share_bookmark=content.get('SHARE_BOOKMARK'),
                    share_bot=content.get('SHARE_BOT'),
                    member_added=content.get('MEMBER_ADDED'),
                    user_mention=content.get('USER_MENTION'),
                )
            return temp
        return None

    @property
    def ai_settings(self) -> AiSettings | None:
        """AI settings of the application."""
        if not self._ai_settings:
            self.fetch('ai_settings')
        if self._ai_settings:
            return Application.AiSettings.from_dict(source=self._ai_settings)
        return None

    @property
    def auth_modes(self) -> AuthModes | None:
        """Authentication modes of the application."""
        if not self._auth_modes:
            self.fetch('auth_modes')
        if self._auth_modes:
            return Application.AuthModes.from_dict(source=self._auth_modes)
        return None

    @property
    def environments(self) -> Environments | None:
        """Environments of the application."""
        if not self._environments:
            self.fetch('environments')
        if self._environments:
            return Application.Environments.from_dict(source=self._environments)
        return None



from typing import TYPE_CHECKING

from mstrio.api import monitors as monitors_api

if TYPE_CHECKING:
    from mstrio.connection import Connection


def get_contents_caches_loop(
    connection: 'Connection',
    project_id: str,
    node: str,
    offset: int = 0,
    limit_per_request: int = 1000,
    status: str | None = None,
    content_type: str | None = None,
    content_format: str | None = None,
    size: str | None = None,
    owner: str | None = None,
    expiration: str | None = None,
    last_updated: str | None = None,
    hit_count: str | None = None,
    sort_by: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> list[dict]:
    response = monitors_api.get_contents_caches(
        connection,
        project_id,
        node,
        offset,
        limit_per_request,
        status,
        content_type,
        content_format,
        size,
        owner,
        expiration,
        last_updated,
        hit_count,
        sort_by,
        fields,
        error_msg,
    )
    caches = response.json()
    total_caches = caches.get('total')
    all_caches = caches.get('contentCaches')
    offset = 0
    while len(all_caches) < total_caches:
        offset += limit_per_request
        response = monitors_api.get_contents_caches(
            connection,
            project_id,
            node,
            offset,
            limit_per_request,
            status,
            content_type,
            content_format,
            size,
            owner,
            expiration,
            last_updated,
            hit_count,
            sort_by,
            fields,
            error_msg,
        )
        caches = response.json()
        all_caches += caches.get('contentCaches')
    return all_caches



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting Runtimes.")
def list_runtimes(
    connection: Connection, fields: str | None = None, error_msg: str | None = None
) -> Response:
    """Get list of runtimes.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """

    return connection.get(endpoint='/api/runtimes', params={'fields': fields})


@ErrorHandler(err_msg="Error creating Runtime.")
def create_runtime(
    connection: Connection,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a runtime.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new runtime
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """

    return connection.post(
        endpoint='/api/runtimes', json=body, params={'fields': fields}
    )


@ErrorHandler(err_msg="Error deleting Runtime with ID: {id}")
def delete_runtime(
    connection: Connection, id: str, error_msg: str | None = None
) -> Response:
    """Delete a runtime.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the runtime to be deleted
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """

    return connection.delete(endpoint=f'/api/runtimes/{id}')



"""This is the demo script to show how to manage devices.
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import (
    BackupPrinterProperties,
    ConnectionParameters,
    Device,
    DeviceType,
    EmailDeviceProperties,
    EmailFormat,
    EmailSmartHostSettings,
    FileDeviceProperties,
    FileLocation,
    FileProperties,
    FileSystem,
    IOSDeviceProperties,
    list_devices,
    PrinterDeviceProperties,
    PrinterLocation,
    PrinterProperties,
    PrinterPdfSettings,
    Transmitter,
    UnixWindowsSharity
)
from mstrio.connection import get_connection

# Define a variable which will be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which will be later used in a script
SMART_HOST_SERVER = $smart_host_server  # server name for smart host
SMART_HOST_PORT = $smart_host_port  # server port for smart host

# create email device properties object
# which is used when creating device
edp = EmailDeviceProperties(
    format=EmailFormat.UU_ENCODED,
    smart_host_settings=EmailSmartHostSettings(server=SMART_HOST_SERVER, port=SMART_HOST_PORT),
)
# see distribution_services/device/device_properties.py for EmailFormat values

# Define variables which can be later used in a script
FEEDBACK_SERVICE_SERVER = $feedback_service_server  # server name for ios feedback server
FEEDBACK_SERVICE_PORT = $feedback_service_port  # server port for ios feedback server

# create ios device properties object
# which is used when creating iphone or ipad device
idp = IOSDeviceProperties(
    app_id='com.microstrategy.dossier.iphone',
    server='api.push.apple.com',
    port=443,
    feedback_service_server=FEEDBACK_SERVICE_SERVER,
    feedback_service_port=FEEDBACK_SERVICE_PORT,
)

# Define a variable which can be later used in a script
FILE_DEVICE_PATH = $file_device_path  # path for file device

# create file device properties object
# which is used when creating file device
file_location = FileLocation(file_path=FILE_DEVICE_PATH)
file_system = FileSystem(create_folder=False)
fp = FileDeviceProperties(
    file_location=file_location,
    file_system=file_system,
    connection_parameters=ConnectionParameters(retries_count=6),
    file_properties=FileProperties(read_only=False),
    unix_windows_sharity=UnixWindowsSharity(sharity_enabled=False),
)

# Define a variable which can be later used in a script
PRINTER_LOCATION = $printer_location  # printer location field for printer device properties

# create print device properties object
# which is used when creating print device
printer_properties = PrinterProperties(pdf_setting=PrinterPdfSettings(post_script_level=1))
pdp = PrinterDeviceProperties(
    printer_location=PrinterLocation(location=PRINTER_LOCATION),
    printer_properties=printer_properties,
    connection_parameters=ConnectionParameters(retries_count=6),
    backup_printer_properties=BackupPrinterProperties(print_on_backup=True),
)

# Define a variable which can be later used in a script
TRANSMITTER_ID = $transmitter_id  # id for Transmitter object

# get transmitters by id
transmitter = Transmitter(conn, id=TRANSMITTER_ID)

# Define variables which can be later used in a script
DEVICE_NAME = $device_name
DEVICE_DESCRIPTION = $device_description

# create a device with device type as `email` (when type is `email` then
# it is mandatory to provide `email_device_properties`)
new_email_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.EMAIL,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=edp,
)
# create a device with device type as `iphone` (when type is `iphone` then
# it is mandatory to provide `device_properties`). Device type
# can be also passed as a string and `device_properties` can be
# passed as a dictionary
new_iphone_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.IPHONE,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=idp,
)

# create a device with device type as `ipad`
new_ipad_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.IPAD,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=idp,
)

# create a device with device type as `file`
new_file_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.FILE,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=fp,
)

# create a device with device type as `printer`
new_printer_device = Device.create(
    connection=conn,
    name=DEVICE_NAME,
    # the DeviceType values can be found in distribution_services/device/device.py
    device_type=DeviceType.PRINTER,
    description=DEVICE_DESCRIPTION,
    transmitter=transmitter,
    device_properties=pdp,
)

# get list of devices
print(list_devices(conn))

# Define a variable which can be later used in a script
DEVICE_ID = $device_id  # id for Device object

# get device by id
device = Device(conn, id=DEVICE_ID)

# get device by name
device_by_name = Device(conn, name=DEVICE_NAME)

# alter email type device and its properties
edp.smart_host_settings.server = 'microstrategy.com'
edp.smart_host_settings.port = 21
new_email_device.alter(
    name=DEVICE_NAME,
    description=DEVICE_DESCRIPTION,
    device_properties=edp,
)

# list device properties
device.list_properties()

# Delete device. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
new_email_device.delete(force=True)
new_iphone_device.delete(force=True)



# flake8: noqa
# isort: off
from .package import (
    Action,
    PackageConfig,
    PackageContentInfo,
    PackageSettings,
    MigrationPurpose,
    PackageType,
    PackageStatus,
    ImportStatus,
)
from .migration import Migration, list_migrations, list_migration_possible_content

# isort: on



"""
    This example is for a dashboard user to modify data in a PostgreSQL DB table using
    a modern grid configured with transaction functions. The workflow will rely on this
    script to make actual modifications (insert, update or delete) in the table.
"""

"""
    The table named USERS has 5 columns USER_ID(Int), USER_LOGIN(Varchar),
    HIRED_DATE(Date), EMAIL(VarChar) and MODIFIED_BY(VarChar), with USER_ID as
    the primary. Data changes for HIRED_DATE(Date), EMAIL(VarChar)
    and MODIFIED_BY(VarChar) are optional.

    In this template we use the library 'psycopg2' (installed in runtime as
    'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests
    and execute operations. For using other types of database, make sure to use
    their methods correctly.

    Python libraries to handle SQL connections are not installed on the environment by
    default so it is necessary to choose runtime which have such library installed in
    the settings of the script. Additionally it is necessary to specify correct `Network
    Access` to allow script to communicate with some other IP then IP of the
    environment.
"""

import psycopg2
import datetime


# helper method to handle not adding None values to the query
def append_not_none_value(array, value):
    if value is not None:
        array.append(value)
        return True
    return False

# Helper method to return datetime object in correct format for SQL operations
def parseDateTimeValue(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000,
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%m/%d/%Y",
        "%m/%d/%Y %H:%M:%S",
        "%Y-%d-%m %H:%M:%S",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m-%d-%Y"
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.datetime.strptime(dateValue, date_format)
            break
        except:
            pass

    if (parsed_datetime == None):
        raise Exception("Wrong Type of the datetime input value conversion!!!!")
    else:
        return parsed_datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method. When providing
    # passwords or any other sensitive data it is good to use variable of type `secret`.
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare queries for database operations. Due to the fact that for some rows
    # we might want to pass `None` value we need some additional logic. This logic
    # require to create query during the time of reading passed properties.

    # Parameterized query for insert new row
    # We assume that we must pass `USER_ID` and `USER_LOGIN` when inserting
    # and the other properties are optional.
    tmp_sql_insert_query = """
        INSERT INTO USERS (
            USER_ID,
            USER_LOGIN
    """

    # Parameterized query for edit the existing row
    # We assume that when updating we must pass only `USER_LOGIN` and the other
    # properties are optional. Update is done based on `USER_ID` and `WHERE` clause
    # is added in the end of preparing query for each row
    tmp_sql_update_query = """
        UPDATE USERS SET USER_LOGIN = %s
    """

    # Parameterized query for select the existing row to delete
    sql_delete_query = """DELETE FROM USERS WHERE USER_ID = %s"""

    # create cursor to execute operations on database
    cursor = connection.cursor()

    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`. Values for `txn_actions` are calculated
    # by dashboard logic and they are based on the action which was created on
    # particular row.
    for index in range(len($txn_actions)):
        if $txn_actions[index] == "insert":
            # when value is optional and not provided, then don't add it to insert query
            optional_values_array = []
            optional_values_in_insert_query = ''
            # this value depends on number of the properties in initial INSERT query
            # it is equal to the number of properties there
            i = 2
            # when array with values for each property is not `None` and the value
            # for particular row is not `None` then we add it
            if $hired_date and $hired_date[index]:
                optional_values_array.append(parseDateTimeValue($hired_date[index]))
                optional_values_in_insert_query += f",HIRED_DATE"
                i = i + 1
            if $email and append_not_none_value(optional_values_array, $email[index]):
                optional_values_in_insert_query += f",EMAIL"
                i = i + 1
            # `modified_by` is a standard variable of type `system_prompt` so its value
            # is calculated based on the chosen System Prompt
            if append_not_none_value(optional_values_array, $modified_by):
                optional_values_in_insert_query += f",MODIFIED_BY"
                i = i + 1

            # create list of parameters
            query_parameters = ",".join(list(map(lambda x: "%s", list(range(i)))))
            optional_values_in_insert_query += f") VALUES ({query_parameters})"

            sql_insert_tuple = (
                int($user_id[index]),
                $user_login[index],
            ) + tuple(optional_values_array)
            # final INSERT query is created based on initial INSERT query and the part
            # created for particular row based on properties passed for this row
            sql_insert_query = tmp_sql_insert_query + optional_values_in_insert_query

            # execute insert query
            cursor.execute(sql_insert_query, sql_insert_tuple)

        elif $txn_actions[index] == "update":
            optional_values_array = []
            optional_values_in_update_query = ''
            # when array with values for each property is not `None` and the value
            # for particular row is not `None` then we add it
            if $hired_date and $hired_date[index]:
                optional_values_array.append(parseDateTimeValue($hired_date[index]))
                optional_values_in_update_query += f",HIRED_DATE = %s"
            if $email and append_not_none_value(optional_values_array, $email[index]):
                optional_values_in_update_query += f",EMAIL = %s"
            # `modified_by` is a standard variable of type `system_prompt` so its value
            # is calculated based on the chosen System Prompt
            if append_not_none_value(optional_values_array, $modified_by):
                optional_values_in_update_query += f",MODIFIED_BY = %s"

            # final UPDATE query is created based on initial UPDATE query and the part
            # created for particular row based on properties passed for this row
            # after adding optional properties we need to add `WHERE` clause
            # which in this case is based on `USER_ID`
            optional_values_array.append(int($user_id[index]))
            optional_values_in_update_query += f" WHERE USER_ID = %s"

            sql_update_tuple = (
                $user_login[index],
            ) + tuple(optional_values_array)
            sql_update_query = tmp_sql_update_query + optional_values_in_update_query
            # execute update query
            cursor.execute(sql_update_query, sql_update_tuple)

        elif $txn_actions[index] == "select":
            # case with DELETE query is the easiest one because we are deleting
            # just based on `USER_ID`
            cursor.execute(sql_delete_query, tuple([int($user_id[index])]))

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()



from time import sleep, time

from requests import Response

from mstrio.api import reports as reports_api
from mstrio.connection import Connection


def get_report_status(
    connection: 'Connection',
    report_id: str,
    instance_id: str,
    project_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
    timeout: int = 60,
) -> Response:
    """Get stable status of a report instance.

    Args:
        connection: Strategy connection object returned by
            `connection.Connection()`.
        report_id (str): ID of the report.
        instance_id (str): ID of the report instance.
        project_id (str, optional): ID of the project.
        fields (str, optional): Comma-separated list of fields to be returned
            in the response.
        error_msg (str, optional): Custom error message to be displayed in case
            of an error.
        timeout (int): Maximum time in seconds to wait for the operation to
            complete. Defaults to 60.

    Returns:
        Complete HTTP response object. Status code 200 when report is ready.
        If timeout is exceeded, returns the last response as-is.
    """

    start_time = time()

    response = reports_api.get_report_status(
        connection=connection,
        report_id=report_id,
        instance_id=instance_id,
        project_id=project_id,
        fields=fields,
        error_msg=error_msg,
    )

    while response.ok and response.status_code == 202:
        # Check if timeout has been exceeded
        elapsed_time = time() - start_time
        if elapsed_time >= timeout:
            return response  # Return the current response instead of raising error

        retry_after = int(response.headers.get('Retry-After', 1))
        sleep(retry_after)

        response = reports_api.get_report_status(
            connection=connection,
            report_id=report_id,
            instance_id=instance_id,
            project_id=project_id,
            fields=fields,
            error_msg=error_msg,
        )

    return response



import requests

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error while creating the package holder")
def create_package_holder(
    connection: Connection,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Create a new in-memory metadata package holder.

    Args:
        connection (Connection): Strategy One REST API connection object.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.post(
        endpoint='/api/packages',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error while updating the package holder with id: {id}")
def update_package_holder(
    connection: Connection,
    body: dict,
    id: str,
    project_id: str | None = None,
    prefer: str = 'respond-async',
    error_msg: str | None = None,
) -> requests.Response:
    """Fill the content of the in-memory metadata package holder per supplied
    specification. Currently, it's only supported when the holder is empty.

    Args:
        connection (Connection): Strategy One REST API connection object.
        body (dict): dictionarized PackageConfig object (with `to_dict()`)
        id (str): ID of the package to be updated
        prefer (str, optional): API currently just supports asynchronous mode,
        not support synchronous mode, so header parameter `Prefer` must be set
        to `respond-async` in your request. Defaults to "respond-async".
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.put(
        endpoint=f'/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        json=body,
    )


@ErrorHandler(err_msg="Error while downloading the package with id: {id}")
def download_package(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Download a package binary.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): ID of the package to be downloaded.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.get(
        endpoint=f'/api/packages/{id}/binary',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error while uploading the package with id: {id}")
def upload_package(
    connection: Connection,
    id: str,
    file: bytes,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Upload package to sandbox directly.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): ID of the package to be uploaded.
        file (bytes): package in a format of a binary string.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.put(
        endpoint=f'/api/packages/{id}/binary',
        headers={'X-MSTR-ProjectID': project_id},
        files={'file': file},
    )


@ErrorHandler(err_msg="Error while getting the package holder with id: {id}")
def get_package_holder(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    show_content: bool = True,
    error_msg: str | None = None,
) -> requests.Response:
    """Get definition of a package, including package status and its detail
    content.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): ID of the package to be retrieved.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        show_content (bool, optional): Show package content or not. Defaults to
        False.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.get(
        endpoint=f'/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'showContent': show_content},
    )


@ErrorHandler(err_msg="Error while deleting the package holder with id: {id}")
def delete_package_holder(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    prefer: str = 'respond-async',
    error_msg: str | None = None,
) -> requests.Response:
    """Delete the in-memory metadata package holder, releasing associated
    Intelligence Server resources.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): ID of the package to be deleted.
        prefer (str, optional): API currently just supports asynchronous mode,
        not support synchronous mode, so header parameter `Prefer` must be set
        to `respond-async` in your request. Defaults to "respond-async".
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.delete(
        endpoint=f'/api/packages/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
    )


@ErrorHandler(
    err_msg="Error while creating the import for package holder with id: {id}"
)
def create_import(
    connection: Connection,
    id: str,
    prefer: str = 'respond-async',
    project_id: str | None = None,
    generate_undo: bool = False,
    error_msg: str | None = None,
) -> requests.Response:
    """Create a package import process.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): ID of the package for which import process will be
            created.
        prefer (str, optional): Allow client to set preferences. Currently,
            respond-async is the only supported mode.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        generate_undo (bool, optional): Generate undo package or not. Defaults
            to False.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.post(
        endpoint='/api/packages/imports',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'packageId': id, 'generateUndo': generate_undo},
    )


@ErrorHandler(err_msg="Error while getting the import with id: {id}")
def get_import(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Get result of a package import process.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Import process ID.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.get(
        endpoint=f'/api/packages/imports/{id}',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error while deleting the import with id: {id}")
def delete_import(
    connection: Connection,
    id: str,
    prefer: str = 'respond-async',
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Closes an existing import process previously created.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Import process ID.
        prefer (str, optional): Allow client to set preferences. Currently,
            respond-async is the only supported mode.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.delete(
        endpoint=f'/api/packages/imports/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
    )


@ErrorHandler(err_msg="Error while creating the undo for import with id: {id}")
def create_undo(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Download undo package binary for this import process.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Import process ID.
        project_id (Optional[str]): Optional ID of a project. Defaults to None.
        error_msg (Optional[str]): Optional error message. Defaults to None.

    Returns:
        requests.Response: Response object containing all of the information
        returned by the server.
    """
    return connection.get(
        endpoint=f'/api/packages/imports/{id}/undoPackage/binary',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error getting migrations list")
def list_migrations(
    connection: Connection,
    offset: int = 0,
    limit: int = 100,
    migration_type: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Get a list of migrations.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        offset (int, optional): Starting point within the collection of returned
            results
        limit (int, optional): Maximum number of items returned for a single
            request. Maximum value: 1000
        migration_type (str, optional): The purpose of the migration package,
            the allowable values include object_migration, project_merge,
            migration_from_shared_file_store. If None it returns only
            object_migration migrations
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(
        endpoint='/api/migrations',
        params={
            'offset': offset,
            'limit': limit,
            'packageInfo.purpose': migration_type,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error getting migration with ID: {migration_id}")
def get_migration(
    connection: Connection,
    migration_id: str,
    show_content: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Get definition of a single metric by id

    Args:
        connection (Connection): Strategy One REST API connection object
        migration_id (str): ID of the migration object
        show_content (str, optional): The type of showing content, the allowable
            values include default, tocview, treeview, all
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    return connection.get(
        endpoint=f'/api/migrations/{migration_id}',
        params={
            'showContent': show_content,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error deleting migration package with ID: {package_id}")
def delete_migration(
    connection,
    package_id,
    error_msg: str | None = None,
    whitelist=[('ERR006', 404)],  # noqa: B006
) -> requests.Response:
    """Delete migration package specified by ID.

    Args:
        connection (Connection): Strategy One REST API connection object
        package_id (str): The ID of the migration package to be deleted
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    return connection.delete(
        endpoint='/api/migrations',
        params={'packageId': package_id},
    )


@ErrorHandler(
    err_msg="Error downloading migration package binary with ID: {package_id}"
)
def download_migration_package(
    connection: Connection,
    package_id: str,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Get definition of a single metric by id

    Args:
        connection (Connection): Strategy One REST API connection object
        package_id (str): The ID of the migration package to be downloaded
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 200.
    """
    return connection.get(
        endpoint=f'/api/migrations/packages/{package_id}/binary',
        params={
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error while creating new migration")
def create_new_migration(
    connection: Connection,
    body: dict,
    project_id: str | None = None,
    prefer: str | None = None,
    unlock_on_start: bool | None = True,
    package_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Create total new migration or reuse existing package to create new
        migration.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server
        body (dict): Migration creation data
        project_id (str, optional): Optional ID of a project. Defaults to None
        prefer (str, optional): Prefer--allow client to set preferences.
            Currently, respond-async allows client to execute in async mode
        unlock_on_start (bool, optional): Whether unlock should be done when
            creation is started. Defaults to True.
        package_id (str, optional): ID of the migration package to be reused
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 201.
    """
    return connection.post(
        endpoint='/api/migrations',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={
            'unlockOnStart': unlock_on_start,
            'packageId': package_id,
            'fields': fields,
        },
        json=body,
    )


@ErrorHandler(err_msg="Error importing migration with ID: {migration_id}")
def start_migration(
    connection: Connection,
    prefer: str,
    migration_id: str,
    body: dict,
    project_id: str | None = None,
    generate_undo: bool | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Trigger an import process to migrate the package from the source
    environment to the target environment.

    Args:
        connection (Connection): Object representation of connection to
            MSTR Server
        prefer (str): Prefer--allow client to set preferences. Currently,
            respond-async allows client to execute in async mode
        migration_id (str): ID of the migration object
        body (dict): Information about the import process
        project_id (str, optional): Optional ID of a project. Defaults to None
        generate_undo (bool, optional): Generate undo package or not
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 201.
    """
    return connection.put(
        endpoint=f'/api/migrations/{migration_id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'generateUndo': generate_undo, 'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error updating or reverting a migration with ID: {migration_id}")
def update_migration(
    connection: Connection,
    migration_id: str,
    body: dict,
    project_id: str | None = None,
    prefer: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Update a migration or trigger an undo process.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        migration_id (str): ID of the migration object to be updated or reverted
        body (dict): Migration information to be updated
        project_id (str, optional): Optional ID of a project. Defaults to None
        prefer (str): Prefer--allow client to set preferences. Currently,
            respond-async allows client to execute in async mode
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected statuses: 200, 202.
    """
    return connection.patch(
        endpoint=f'/api/migrations/{migration_id}',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'migrationId': migration_id, 'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error editing migration with ID: {migration_id}")
def edit_migration_package(
    connection: Connection,
    prefer: str,
    migration_id: str,
    body: dict,
    project_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Edit migration package information.

    Args:
        connection (Connection): Strategy One REST API connection object
        prefer (str): Prefer--allow client to set preferences. Currently,
            respond-async allows client to execute in async mode
        migration_id (str): ID of the migration object
        body (dict): Migration information to be updated
        project_id (str, optional): Optional ID of a project. Defaults to None
        fields(str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object. Expected status: 202.
    """
    return connection.put(
        endpoint=f'/api/migrations/{migration_id}/packageInfo',
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error creating file metadata.")
def storage_service_create_file_metadata(
    connection: Connection,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Create a file metadata before uploading the file, which is stored in the
    storage service.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): {
            "name": str,
            "extension": str ("mmp"),
            "type": str ("migrations.packages"),
            "size": int,
            "sha256": str,
            "environment": {
                "id": str,
                "name": str,
            },
            "extraInfo": {
                "packageId": str (optional),
                "packageType": str
                    ("project", "project_security", "configuration"),
                "packagePurpose": str
                    ("object_migration", "project_merge", "migration_group",
                    "migration_group_child", "migration_from_shared_file_store")
            },
        }
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.post(
        endpoint='/api/mstrServices/library/storage/sharedFileStore/files',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error downloading file binary for ID {file_id}.")
def storage_service_download_file_binary(
    connection: Connection,
    file_id: str,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Upload a file binary to storage service directly.

    Args:
        connection (Connection): Strategy One REST API connection object
        file_id (str): file ID
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success, with MIME type
            application/octet-stream.
    """
    return connection.get(
        endpoint=(
            f'/api/mstrServices/library/storage/sharedFileStore/files/{file_id}/binary'
        ),
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error uploading file binary for ID {file_id}.")
def storage_service_upload_file_binary(
    connection: Connection,
    file_id: str,
    file: bytes,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Upload a file binary to storage service directly.

    Args:
        connection (Connection): Strategy One REST API connection object
        file_id (str): file ID
        file (bytes): file object
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.put(
        endpoint=(
            f'/api/mstrServices/library/storage/sharedFileStore/files/{file_id}/binary'
        ),
        params={'fields': fields},
        files={'file': file},
    )


@ErrorHandler(err_msg="Error deleting file binary with ID: {file_id}.")
def delete_file_binary(
    connection: Connection,
    file_id: str,
    error_msg: str | None = None,
) -> requests.Response:
    """Delete a file binary from storage service directly.

    Args:
        connection (Connection): Strategy One REST API connection object
        file_id (str): file ID to be deleted
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    return connection.delete(
        endpoint=(f'/api/mstrServices/library/storage/sharedFileStore/files/{file_id}')
    )


@ErrorHandler(err_msg="Error getting storage files list")
def list_storage_files(
    connection: Connection,
    file_type: str,
    offset: int = 0,
    limit: int = 100,
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Get a list of storage files.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        offset (int, optional): Starting point within the collection of returned
            results
        limit (int, optional): Maximum number of items returned for a single
            request. Maximum value: 1000
        file_type (str, optional): The type of shared file which you want to
            query. Currently, only file type "migration.packages" is supported
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.get(
        endpoint='/api/mstrServices/library/storage/sharedFileStore/files',
        params={
            'offset': offset,
            'limit': limit,
            'type': file_type,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error certifying migration package with ID: {id}")
def certify_migration_package(
    connection: Connection,
    id: str,
    body: dict | None = None,
    auto_sync: bool | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Update a migration's package certification status or trigger a process
    to synchronize the status.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): migration package ID
        body (dict, optional): Migration package information to be updated
        auto_sync (bool, optional): Whether to trigger a process to synchronize
            the certification status with shared environment via storage service
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    return connection.patch(
        endpoint=f'/api/migrations/{id}/packageInfo/certification',
        json=body,
        params={'autoSync': auto_sync},
    )


@ErrorHandler(
    err_msg="Error triggering validation process for migration package with ID: {id}"
)
def trigger_migration_package_validation(
    connection: Connection,
    id: str,
    body: dict,
    project_id: str | None = None,
    prefer: str = 'respond-async',
    fields: str | None = None,
    error_msg: str | None = None,
) -> requests.Response:
    """Trigger a validation process

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): migration package ID
        body (dict): Validation information
        project_id (str, optional): ID of the project
        prefer (str, optional): Prefer--allow client to set preferences.
            Currently, respond-async allows client to execute in async mode
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 202.
    """
    return connection.put(
        endpoint=f'/api/migrations/{id}/validation',
        json=body,
        headers={'X-MSTR-ProjectID': project_id, 'Prefer': prefer},
        params={'fields': fields},
    )



"""This is the demo script to show how to manage dynamic recipient lists. 
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import get_connection
from mstrio.distribution_services.subscription import DynamicRecipientList, list_dynamic_recipient_lists

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define variables which can be later used in a script
# DRL stands for Dynamic Recipient List
DRL_NAME = $drl_name
SOURCE_REPORT_ID = $source_report_id
ATTRIBUTE_ID = $attribute_id
ATTRIBUTE_FORM_ID = $attribute_form_id
DESCRIPTION = $description
EXAMPLE_MAPPING_FIELD = DynamicRecipientList.MappingField(
    attribute_id=ATTRIBUTE_ID, 
    attribute_form_id=ATTRIBUTE_FORM_ID)

# Create a DRL
drl3 = DynamicRecipientList.create(connection=conn, name = DRL_NAME, description=DESCRIPTION,
    source_report_id=SOURCE_REPORT_ID,
    physical_address=EXAMPLE_MAPPING_FIELD, 
    linked_user=EXAMPLE_MAPPING_FIELD, 
    device=EXAMPLE_MAPPING_FIELD)

# List dynamic recipient lists with different conditions
list_of_all_drl = list_dynamic_recipient_lists(connection=conn)
print(list_of_all_drl)
list_of_all_drl_as_dicts = list_dynamic_recipient_lists(connection=conn, to_dictionary=True)
print(list_of_all_drl_as_dicts)
list_of_all_drl_by_name = list_dynamic_recipient_lists(connection=conn, name=DRL_NAME)
print(list_of_all_drl_by_name)

# Define variables which can be later used in a script
DRL_ID = $drl_id # ID of the DRL we want to perform actions on

# Get a DRL by it's id or name
drl = DynamicRecipientList(connection=conn, id=DRL_ID)
drl2 = DynamicRecipientList(connection=conn, name=DRL_NAME)

# List properties of a DRL
print(drl.list_properties())

# Alter a DRL
drl2.alter(name = DRL_NAME, description=DESCRIPTION,
    source_report_id=SOURCE_REPORT_ID,
    physical_address=EXAMPLE_MAPPING_FIELD,
    linked_user=EXAMPLE_MAPPING_FIELD, 
    device=EXAMPLE_MAPPING_FIELD)

# Delete a DRL without prompt
drl.delete(force=True)



# x is a scalar input
# y is a scalar input
# Metric expression example: ScriptFunctionSimple<_InputNames="x, y", _ScriptName=”minus_two_numbers”>(Revenue, Cost)

def minus_two_numbers(x, y):
    return x - y

# must-have method which returns the values served as the metric evaluation results. 
def get_results():
    return minus_two_numbers($x, $y)


print(get_results())



import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from typing import Optional

from mstrio.connection import Connection
from mstrio.server import Environment
from mstrio.server.project import Project
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, exception_handler
from mstrio.utils.time_helper import DatetimeFormats

logger = logging.getLogger(__name__)


class RequestStatus(AutoName):
    """Status of a request such as import or undo.
    Allowed values are:
    - unknown
    - pending
    - requested
    - rejected
    - approved
    """

    UNKNOWN = auto()
    PENDING = auto()
    REQUESTED = auto()
    REJECTED = auto()
    APPROVED = auto()


class Action(AutoName):
    """The default action used for objects which don't have actions
    explicitly, for example the dependents objects.
    USE_EXISTING: No change is made to the destination object.
    The source object is not copied.
    REPLACE: The destination object is replaced with the source object.
    Note the following:
        • If the conflict type is Exists identically except for path, or
        Exists identically except for Distribution Services objects,
        the  destination object is updated to reflect the path or
        Distribution Services addresses and contacts of the source object.
        • Replace moves the object into the same parent folder as source
        object. If the parent path is the same between source and
        destination but the grandparent path is different, Replace may
        appear to do nothing because Replace puts the object into the same
        parent path.
        • Non-empty folders in the destination location will never have
        the same version ID and modification time as the source, because
        the folder is copied first and the objects are added to it, thus
        changing version ID and modification times during the copy process.
    KEEP_BOTH: No change is made to the destination object. The source
    object is duplicated if destination object doesn't exist. But if the
    destination object exists with the same id and same version, this source
    object is ignored.If the destination object exists with the same id
    and different version, this source object is saved as new object.
    USE_NEWER: If the source object's modification time is more recent than
    the destination object's, the Replace action is used.Otherwise, the Use
    existing action is used.
    USE_OLDER: If the source object's modification time is more recent than
    the destination object's, the Use existing action is used. Otherwise,
    the Replace action is used.
    FORCE_REPLACE: Replace the object in the destination project with the
    version of the object in the update package, even if both versions of
    the object have the same Version ID.
    DELETE: Delete the object from the destination project. The version
    of the object in the update package is not imported into the destination
    project.Warning: If the object in the destination has any used-by
    dependencies when you import the update package, the import will fail.
    """

    USE_EXISTING = auto()
    REPLACE = auto()
    KEEP_BOTH = auto()
    USE_NEWER = auto()
    USE_OLDER = auto()
    FORCE_REPLACE = auto()
    DELETE = auto()


class TranslationAction(AutoName):
    """Translation action for Project Merge Package Settings.
    Allowed values are:
    - not_merged
    - use_existing
    - replace
    - force_replace
    """

    NOT_MERGED = auto()
    USE_EXISTING = auto()
    REPLACE = auto()
    FORCE_REPLACE = auto()


@dataclass
class PackageStorage(Dictable):
    size: int | None = None
    path: str | None = None


@dataclass
class PackageWarning(Dictable):
    message: str | None = None
    iserver_error_code: int | None = None
    banned_object_count: int | None = None


class PackageSettings(Dictable):
    """Object representation of package settings details.

    Attributes:
        default_action(Action): default action for Package
        update_schema(Optional[UpdateSchema]): update_schema for Package
        acl_on_replacing_objects(Optional[AclOnReplacingObjects]):
            ACL setting on replacing objects
        acl_on_new_objects(Optional[AclOnNewObjects]):
            ACL setting on new objects
    """

    class UpdateSchema(AutoName):
        """They allow you to configure the package to automatically perform
        certain schema update functions. These options can be useful if you make
        any changes to schema objects. Use the recalculate table keys and fact
        entry levels if you changed the key structure of a table or if you
        changed the level at which a fact is stored.Use the recalculate table
        logical sizes to override any modifications that you have made to
        logical table sizes. (Logical table sizes affect how the Strategy One
        SQL Engine determines which tables to use in a query.)
        """

        RECAL_TABLE_LOGICAL_SIZE = auto()
        RECAL_TABLE_KEYS_FACT_ENTRY_LEVEL = auto()
        UPDATE_SCHEMA_LOGICAL_INFO = auto()

    class AclOnReplacingObjects(AutoName):
        """If you resolve a conflict with the "Replace" action, and the access
        control lists (ACL) of the objects are different between the two
        projects, you can choose whether to keep the existing ACL in the
        destination project or replace it with the ACL from the source project.
        Note: This is not supported for project security packages.
        """

        USE_EXISTING = auto()
        REPLACE = auto()
        KEEP_BOTH = auto()

    class AclOnNewObjects(AutoName):
        """If you add a new object to the destination project with the
        "Create New" or "Keep Both action", you can choose to have the object
        inherit its ACL from the destination folder instead of keeping its own
        ACL. This is helpful when copying an object into a user's profile
        folder, so that the user can have full control over the object.
        """

        KEEP_ACL_AS_SOURCE_OBJECT = auto()
        INHERIT_ACL_AS_DEST_FOLDER = auto()

    _FROM_DICT_MAP = {
        'default_action': Action,
        'update_schema': [UpdateSchema],
        'acl_on_replacing_objects': AclOnReplacingObjects,
        'acl_on_new_objects': [AclOnNewObjects],
    }

    def __init__(
        self,
        default_action: Action = Action.USE_EXISTING,
        update_schema: UpdateSchema | None = None,
        acl_on_replacing_objects: AclOnReplacingObjects = (
            AclOnReplacingObjects.USE_EXISTING
        ),
        acl_on_new_objects: AclOnNewObjects = AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
    ):
        self.default_action = default_action
        self.update_schema = (
            update_schema
            if isinstance(update_schema, list) or not update_schema
            else [update_schema]
        )
        self.acl_on_replacing_objects = acl_on_replacing_objects
        self.acl_on_new_objects = (
            [acl_on_new_objects]
            if isinstance(acl_on_new_objects, PackageSettings.AclOnNewObjects)
            else acl_on_new_objects
        )


class PackageContentInfo(Dictable):
    """Object representation of package content information

    Attributes:
        id(str): object ID
        action(Action): The action to resolve the conflict.
        name(Optional[str]): object name
        version(Optional[str]): object version
        type(Optional[ObjectTypes]): object type
        owner(Optional[Owner]): owner of object
        date_created(Optional[str]): object creation date
        date_modified(Optional[str]): object modification date
        include_dependents(Optional[bool]): whether include the dependents
            for this object or not
        explicit_included(Optional[bool]): whether explicitly included or not
        level(Optional[Level]): level of object
    """

    class Level(AutoName):
        PROJECT_OBJECT = auto()
        CONFIGURATION_OBJECT = auto()
        UNKNOWN = auto()

    class Owner(User):
        """
        This class overrides `from_dict()` and `__init__()` methods as
        PackageContentInfo inherits from `Dictable` and not `EntityBase`.
        Therefore, despite `Owner` being a `User`, `connection` param here is
        optional. If `connection` is specified, `User` constructor will be used
        """

        @classmethod
        def from_dict(
            cls, source: dict[str, any], connection: Optional["Connection"] = None
        ):
            return super().from_dict(source, connection)

        def __init__(self, id: str, connection: Optional["Connection"] = None):
            if connection:
                super().__init__(id, connection)
            else:
                self._id = id

        def _init_variables(self, **kwargs) -> None:
            self._connection = kwargs.get("connection")
            self._id = kwargs.get("id")

    _FROM_DICT_MAP = {
        'type': ObjectTypes,
        'action': Action,
        'level': Level,
        'owner': Owner,
    }

    def __init__(
        self,
        id: str,
        action: Action | str = Action.USE_EXISTING,
        name: str | None = None,
        version: str | None = None,
        type: ObjectTypes | None = None,
        owner: Owner | None = None,
        date_created: str | None = None,
        date_modified: str | None = None,
        include_dependents: bool | None = None,
        explicit_included: bool | None = None,
        level: Level | str | None = None,
    ):
        self.id = id
        self.name = name
        self.version = version
        self.type = type
        self.owner = owner
        self.date_created = date_created
        self.date_modified = date_modified
        self.include_dependents = include_dependents
        self.explicit_included = explicit_included
        try:
            self.level = (
                PackageContentInfo.Level(level) if isinstance(level, str) else level
            )
            self.action = Action(action) if isinstance(action, str) else action
        except ValueError:
            exception_handler(msg="Wrong enum value", exception_type=ValueError)


@dataclass
class ObjectMigrationPackageTocView(Dictable):
    settings: PackageSettings
    content: list[PackageContentInfo] | None = None


@dataclass
class PackageObjectAction(Dictable):
    id: str
    action: Action
    object_type: ObjectTypes | None = None


@dataclass
class PackageObjectTypeAction(Dictable):
    type: ObjectTypes
    action: Action
    sub_type: ObjectSubTypes | None = None


@dataclass
class ProjectMergePackageSettings(Dictable):
    acl_on_replacing_objects: PackageSettings.AclOnReplacingObjects
    acl_on_new_objects: PackageSettings.AclOnNewObjects
    default_action: Action
    validate_dependencies: bool
    translation_action: TranslationAction
    folder_actions: dict | None = None
    object_actions: dict | None = None
    object_type_actions: dict | None = None
    update_schema: PackageSettings.UpdateSchema | None = None


@dataclass
class ProjectMergePackageTocView(Dictable):
    settings: ProjectMergePackageSettings


class PackageConfig(Dictable):
    """Package Update Data Transfer Object

    Attributes:
        settings(PackageSettings): settings details of package
        content(PackageContentInfo, List[PackageContentInfo]): content details
            of package
    """

    _FROM_DICT_MAP = {
        'settings': PackageSettings,
        'content': [PackageContentInfo],
    }

    def __init__(
        self,
        settings: PackageSettings,
        content: list[PackageContentInfo] | PackageContentInfo,
    ):
        self.settings = settings
        self.content = [content] if not isinstance(content, list) else content


class MigrationPurpose(AutoName):
    """Migration purpose.
    Possible values are:
    - object_migration
    - project_merge
    - migration_from_shared_file_store
    """

    OBJECT_MIGRATION = auto()
    PROJECT_MERGE = auto()
    FROM_FILE = 'migration_from_shared_file_store'


class PackageType(Enum):
    """Migration type."""

    OBJECT = 'project'
    ADMINISTRATION = 'configuration'
    PROJECT = 'project_security'


class ImportStatus(AutoName):
    """Status of Migration Import.
    Allowed values are:
    - pending
    - importing
    - imported
    - import_failed
    - undoing
    - undo_success
    - undo_failed
    """

    PENDING = auto()
    IMPORTING = auto()
    IMPORTED = auto()
    IMPORT_FAILED = auto()
    UNDOING = auto()
    UNDO_SUCCESS = auto()
    UNDO_FAILED = auto()


class PackageStatus(AutoName):
    """Status of Migration Package.
    Allowed values are:
    - locked
    - created
    - create_failed
    - creating
    - empty
    - unknown
    """

    LOCKED = auto()
    CREATED = auto()
    CREATE_FAILED = auto()
    CREATING = auto()
    EMPTY = auto()
    UNKNOWN = auto()


class MigratedObjectTypes(AutoName):
    # Per analogous list in Workstation plugin
    DASHBOARD = auto()
    DOCUMENT = auto()
    CARD = auto()
    REPORT = auto()
    DATASET = auto()
    BOT = auto()
    AGENT = "bot"
    ATTRIBUTE = auto()
    BASE_FORMULA = auto()
    CONSOLIDATION = auto()
    CUSTOM_GROUP = auto()
    CONSOLIDATION_ELEMENT = auto()
    COLUMN = auto()
    DRILL_MAP = auto()
    FACT = auto()
    FILTER = auto()
    FUNCTION = auto()
    HIERARCHY = auto()
    METRIC = auto()
    PROMPT = auto()
    SMART_ATTRIBUTE = auto()
    SUBTOTAL = auto()
    AUTO_STYLE = auto()
    EXTERNAL_SHORTCUT = auto()
    SEARCH = auto()
    SECURITY_FILTER = auto()
    SHORTCUT = auto() or 67
    TABLE = auto()
    TEMPLATE = auto()
    TRANSFORMATION = auto()
    SCRIPT = auto()
    USER = auto()
    USER_GROUP = auto()
    DATA_SOURCE = auto()
    SECURITY_ROLE = auto()
    CUSTOM_APPLICATION = auto()
    PALETTE = auto()
    TIMEZONE = auto()
    CALENDAR = auto()
    CONTENT_BUNDLE = auto()
    DEVICE = auto()
    LOCALE = auto()
    SCHEDULE_EVENT = auto()
    SCHEDULE_TRIGGER = auto()
    TRANSMITTER = auto()
    SCRIPT_RUNTIME_ENV = auto()


class PackageCertificationStatus(AutoName):
    UNCERTIFIED = auto()
    REQUESTED = auto()
    CERTIFIED = auto()
    REJECTED = auto()


@dataclass
class PackageCertificationInfo(Dictable):
    _FROM_DICT_MAP = {
        'status': PackageCertificationStatus,
        'operator': User.from_dict,
        'last_updated_date': DatetimeFormats.YMDHMS,
    }
    status: PackageCertificationStatus
    operator: User
    last_updated_date: datetime


class ValidationStatus(AutoName):
    VALIDATING = auto()
    VALIDATED = auto()
    VALIDATION_FAILED = auto()


@dataclass
class Validation(Dictable):
    _FROM_DICT_MAP = {
        'status': ValidationStatus,
        'creation_date': DatetimeFormats.YMDHMS,
        'last_update_date': DatetimeFormats.YMDHMS,
    }
    status: ValidationStatus | None = None
    progress: float | None = None
    message: str | None = None
    creation_date: datetime | None = None
    last_update_date: datetime | None = None


@dataclass
class ImportInfo(Dictable):
    """Python representation of a ImportInfo Migration field.
    Attributes:
        id (str): ID of the import
        creator (dict, optional): Creator data
            id
            name
            full_name (str, optional)
        creation_date (str, optional): Date of creation, in the format
            "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
        last_updated_date (str, optional): Date of last update, in the format
            "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
        environment (dict): Environment data
            id
            name
        status (ImportStatus, optional): Status of the import.
        import_request_status (RequestStatus, optional): Status of the import
            request.
        undo_request_status (RequestStatus, optional): Status of the undo
            request.
        progress (float, optional): Progress of the import.
        message (str, optional): Record information such as the reason of
            import failure
        undo_storage (dict, optional): Storage data for the undo package
            size
            path
        project (dict, optional): Project data
            id
            name
        deleted (bool, optional): Whether the import is deleted
    """

    _FROM_DICT_MAP = {
        'creator': User.from_dict,
        'creation_date': DatetimeFormats.YMDHMS,
        'last_updated_date': DatetimeFormats.YMDHMS,
        'status': ImportStatus,
        'import_request_status': RequestStatus,
        'undo_request_status': RequestStatus,
        'project': Project.from_dict,
        'undo_storage': PackageStorage.from_dict,
    }
    id: str
    environment: Environment | dict
    creator: User | None = None
    creation_date: datetime | None = None
    last_updated_date: datetime | None = None
    status: ImportStatus | None = None
    import_request_status: RequestStatus | None = None
    undo_request_status: RequestStatus | None = None
    progress: float | None = None
    message: str | None = None
    undo_storage: PackageStorage | None = None
    project: Project | None = None
    deleted: bool | None = None


@dataclass
class PackageInfo(Dictable):
    """Python representation of a PackageInfo Migration field.

    Attributes:
        attribute_id: ID of the mapped attribute
        attribute_form_id: ID of the mapped attribute's form
    """

    _FROM_DICT_MAP = {
        'creator': User.from_dict,
        'creation_date': DatetimeFormats.YMDHMS,
        'last_updated_date': DatetimeFormats.YMDHMS,
        'type': PackageType,
        'storage': PackageStorage,
        'project': Project.from_dict,
        'status': PackageStatus,
        'warnings': PackageWarning.bulk_from_dict,
        'purpose': MigrationPurpose,
        'certification': PackageCertificationInfo,
    }
    name: str
    creator: User
    creation_date: datetime
    last_updated_date: datetime
    type: PackageType
    environment: str | None = None
    id: str | None = None
    replicated: bool | None = None
    storage: PackageStorage | None = None
    project: Project | None = None
    status: PackageStatus | None = None
    message: str | None = None
    warnings: list[PackageWarning] | None = None
    progress: float | None = None
    deleted: bool | None = None
    existing: bool | None = None
    toc_view: ProjectMergePackageTocView | ObjectMigrationPackageTocView | None = None
    purpose: MigrationPurpose | None = None
    tree_view: dict | None = None
    certification: PackageCertificationInfo | None = None


OBJECT_MIGRATION_TYPES_MIN_VERSION = {
    MigratedObjectTypes.BOT: '11.3.0760',
    MigratedObjectTypes.AGENT: '11.5.1000',
    MigratedObjectTypes.SMART_ATTRIBUTE: '11.3.0760',
    MigratedObjectTypes.SCRIPT: '11.3.0760',
    MigratedObjectTypes.CALENDAR: '11.3.0760',
    MigratedObjectTypes.CONTENT_BUNDLE: '11.3.0760',
    MigratedObjectTypes.DEVICE: '11.3.0760',
    MigratedObjectTypes.LOCALE: '11.3.0760',
    MigratedObjectTypes.SCHEDULE_EVENT: '11.3.0760',
    MigratedObjectTypes.SCHEDULE_TRIGGER: '11.3.0760',
    MigratedObjectTypes.TRANSMITTER: '11.3.0760',
    MigratedObjectTypes.SCRIPT_RUNTIME_ENV: '11.3.0760',
}
CATALOG_ITEMS = {
    MigratedObjectTypes.DASHBOARD: (ObjectTypes.DOCUMENT_DEFINITION, None),
    MigratedObjectTypes.DOCUMENT: (ObjectTypes.DOCUMENT_DEFINITION, None),
    MigratedObjectTypes.CARD: (
        ObjectTypes.REPORT_DEFINITION,
        ObjectSubTypes.REPORT_HYPER_CARD,
    ),  # Card
    MigratedObjectTypes.REPORT: (
        ObjectTypes.REPORT_DEFINITION,
        [
            ObjectSubTypes.REPORT_GRID,
            ObjectSubTypes.REPORT_GRAPH,
            ObjectSubTypes.REPORT_ENGINE,
            ObjectSubTypes.REPORT_TEXT,
            ObjectSubTypes.REPORT_BASE,
            ObjectSubTypes.REPORT_GRID_AND_GRAPH,
            ObjectSubTypes.REPORT_NON_INTERACTIVE,
            ObjectSubTypes.INCREMENTAL_REFRESH_REPORT,
            ObjectSubTypes.REPORT_TRANSACTION,
            ObjectSubTypes.SUPER_CUBE_IRR,
            ObjectSubTypes.REPORT_DATAMART,
        ],
    ),  # Report
    MigratedObjectTypes.DATASET: (
        ObjectTypes.REPORT_DEFINITION,
        [
            ObjectSubTypes.OLAP_CUBE,
            ObjectSubTypes.SUPER_CUBE,
        ],
    ),  # Dataset
    MigratedObjectTypes.BOT: (
        ObjectTypes.DOCUMENT_DEFINITION,
        ObjectSubTypes.DOCUMENT_BOT,
    ),  # Bot
    MigratedObjectTypes.AGENT: (
        ObjectTypes.DOCUMENT_DEFINITION,
        [
            ObjectSubTypes.DOCUMENT_BOT_2_0,
            ObjectSubTypes.DOCUMENT_BOT_UNIVERSAL,
        ],
    ),  # Agent
    MigratedObjectTypes.ATTRIBUTE: (
        ObjectTypes.ATTRIBUTE,
        [
            ObjectSubTypes.ATTRIBUTE,
            ObjectSubTypes.ATTRIBUTE_ROLE,
            ObjectSubTypes.ATTRIBUTE_TRANSFORMATION,
            ObjectSubTypes.ATTRIBUTE_ABSTRACT,
            ObjectSubTypes.ATTRIBUTE_RECURSIVE,
            ObjectSubTypes.ATTRIBUTE_DERIVED,
        ],
    ),  # Attribute
    MigratedObjectTypes.BASE_FORMULA: (ObjectTypes.AGG_METRIC, None),  # Base Formula
    MigratedObjectTypes.CONSOLIDATION: (
        ObjectTypes.CONSOLIDATION,
        None,
    ),  # Consolidation
    MigratedObjectTypes.CUSTOM_GROUP: (
        ObjectTypes.FILTER,
        ObjectSubTypes.CUSTOM_GROUP,
    ),  # Custom Group
    MigratedObjectTypes.CONSOLIDATION_ELEMENT: (
        ObjectTypes.CONSOLIDATION_ELEMENT,
        None,
    ),  # Consolidation Element
    MigratedObjectTypes.COLUMN: (ObjectTypes.COLUMN, None),  # Column
    MigratedObjectTypes.DRILL_MAP: (ObjectTypes.DRILL_MAP, None),  # Drill Map
    MigratedObjectTypes.FACT: (ObjectTypes.FACT, None),  # Fact
    MigratedObjectTypes.FILTER: (ObjectTypes.FILTER, ObjectSubTypes.FILTER),  # Filter
    MigratedObjectTypes.FUNCTION: (ObjectTypes.FUNCTION, None),  # Function
    MigratedObjectTypes.HIERARCHY: (
        ObjectTypes.DIMENSION,
        [
            ObjectSubTypes.DIMENSION_USER,
            ObjectSubTypes.DIMENSION_ORDERED,
            ObjectSubTypes.DIMENSION_USER_HIERARCHY,
        ],
    ),  # Hierarchy
    MigratedObjectTypes.METRIC: (ObjectTypes.METRIC, ObjectSubTypes.METRIC),  # Metric
    MigratedObjectTypes.PROMPT: (ObjectTypes.PROMPT, None),  # Prompt
    MigratedObjectTypes.SMART_ATTRIBUTE: (
        ObjectTypes.ATTRIBUTE,
        ObjectSubTypes.ATTRIBUTE_SMART,
    ),  # Smart Attribute
    MigratedObjectTypes.SUBTOTAL: (
        ObjectTypes.METRIC,
        [
            ObjectSubTypes.SUBTOTAL_DEFINITION,
            ObjectSubTypes.SYSTEM_SUBTOTAL,
        ],
    ),  # Subtotal
    MigratedObjectTypes.AUTO_STYLE: (ObjectTypes.GRAPH_STYLE, None),  # Auto Style
    MigratedObjectTypes.EXTERNAL_SHORTCUT: (
        ObjectTypes.SHORTCUT,
        None,
    ),  # External Shortcut
    MigratedObjectTypes.SEARCH: (ObjectTypes.SEARCH, None),  # Search
    MigratedObjectTypes.SECURITY_FILTER: (
        ObjectTypes.SECURITY_FILTER,
        None,
    ),  # Security Filter
    MigratedObjectTypes.SHORTCUT: (ObjectTypes.SHORTCUT_TYPE, None),  # Shortcut
    MigratedObjectTypes.TABLE: (ObjectTypes.TABLE, None),  # Table
    MigratedObjectTypes.TEMPLATE: (ObjectTypes.TEMPLATE, None),  # Template
    MigratedObjectTypes.TRANSFORMATION: (
        ObjectTypes.ROLE,
        ObjectSubTypes.ROLE_TRANSFORMATION,
    ),  # Transformation
    MigratedObjectTypes.SCRIPT: (
        ObjectTypes.SCRIPT,
        [
            ObjectSubTypes.SCRIPT,
            ObjectSubTypes.TRANSACTION_SCRIPT,
        ],
    ),  # Script
    MigratedObjectTypes.USER: (ObjectTypes.USER, [ObjectSubTypes.USER]),  # User
    MigratedObjectTypes.USER_GROUP: (
        ObjectTypes.USER,
        [ObjectSubTypes.USER_GROUP],
    ),  # User Group
    MigratedObjectTypes.DATA_SOURCE: (ObjectTypes.DBROLE, None),  # Data Source
    MigratedObjectTypes.SECURITY_ROLE: (
        ObjectTypes.SECURITY_ROLE,
        None,
    ),  # Security Role
    MigratedObjectTypes.CUSTOM_APPLICATION: (
        ObjectTypes.APPLICATION,
        None,
    ),  # Custom Application
    MigratedObjectTypes.PALETTE: (
        ObjectTypes.PALETTE,
        ObjectSubTypes.CUSTOM_PALETTE,
    ),  # Palette
    MigratedObjectTypes.TIMEZONE: (
        ObjectTypes.TIMEZONE,
        [ObjectSubTypes.TIMEZONE_CUSTOM],
    ),  # Timezone
    MigratedObjectTypes.CALENDAR: (
        ObjectTypes.CALENDAR,
        [ObjectSubTypes.CALENDAR_CUSTOM],
    ),  # Calendar
    MigratedObjectTypes.CONTENT_BUNDLE: (
        ObjectTypes.CONTENT_BUNDLE,
        None,
    ),  # Content Bundle
    MigratedObjectTypes.DEVICE: (
        ObjectTypes.SUBSCRIPTION_DEVICE,
        None,
    ),  # Device
    MigratedObjectTypes.LOCALE: (ObjectTypes.LOCALE, None),  # Locale
    MigratedObjectTypes.SCHEDULE_EVENT: (
        ObjectTypes.SCHEDULE_EVENT,
        None,
    ),  # Schedule Event
    MigratedObjectTypes.SCHEDULE_TRIGGER: (
        ObjectTypes.SCHEDULE_TRIGGER,
        None,
    ),  # Schedule Trigger
    MigratedObjectTypes.TRANSMITTER: (
        ObjectTypes.SUBSCRIPTION_TRANSMITTER,
        None,
    ),  # Transmitter
    MigratedObjectTypes.SCRIPT_RUNTIME_ENV: (
        ObjectTypes.RUNTIME,
        None,
    ),  # Script Runtime Environment
}
OBJECT_MIGRATION_TYPES_OBJECT = {
    MigratedObjectTypes.DASHBOARD,
    MigratedObjectTypes.DOCUMENT,
    MigratedObjectTypes.CARD,
    MigratedObjectTypes.REPORT,
    MigratedObjectTypes.DATASET,
    MigratedObjectTypes.BOT,
    MigratedObjectTypes.AGENT,
    MigratedObjectTypes.ATTRIBUTE,
    MigratedObjectTypes.BASE_FORMULA,
    MigratedObjectTypes.CONSOLIDATION,
    MigratedObjectTypes.CUSTOM_GROUP,
    MigratedObjectTypes.CONSOLIDATION_ELEMENT,
    MigratedObjectTypes.COLUMN,
    MigratedObjectTypes.DRILL_MAP,
    MigratedObjectTypes.FACT,
    MigratedObjectTypes.FILTER,
    MigratedObjectTypes.FUNCTION,
    MigratedObjectTypes.HIERARCHY,
    MigratedObjectTypes.METRIC,
    MigratedObjectTypes.PROMPT,
    MigratedObjectTypes.SMART_ATTRIBUTE,
    MigratedObjectTypes.SUBTOTAL,
    MigratedObjectTypes.AUTO_STYLE,
    MigratedObjectTypes.EXTERNAL_SHORTCUT,
    MigratedObjectTypes.SEARCH,
    MigratedObjectTypes.SECURITY_FILTER,
    MigratedObjectTypes.SHORTCUT,
    MigratedObjectTypes.TABLE,
    MigratedObjectTypes.TEMPLATE,
    MigratedObjectTypes.TRANSFORMATION,
    MigratedObjectTypes.SCRIPT,
}
OBJECT_MIGRATION_TYPES_ADMINISTRATION = {
    MigratedObjectTypes.USER,
    MigratedObjectTypes.USER_GROUP,
    MigratedObjectTypes.DATA_SOURCE,
    MigratedObjectTypes.SECURITY_ROLE,
    MigratedObjectTypes.CUSTOM_APPLICATION,
    MigratedObjectTypes.PALETTE,
    MigratedObjectTypes.TIMEZONE,
    MigratedObjectTypes.CALENDAR,
    MigratedObjectTypes.CONTENT_BUNDLE,
    MigratedObjectTypes.DEVICE,
    MigratedObjectTypes.LOCALE,
    MigratedObjectTypes.SCHEDULE_EVENT,
    MigratedObjectTypes.SCHEDULE_TRIGGER,
    MigratedObjectTypes.TRANSMITTER,
    MigratedObjectTypes.SCRIPT_RUNTIME_ENV,
}



from enum import Enum, IntEnum, auto

from mstrio.utils.enum_helper import AutoUpperName


class CertifiedStatus(Enum):
    """Enumeration that represents what can be passed in the certified_status
    attribute of the IServer quick search command."""

    ALL = 'ALL'
    CERTIFIED_ONLY = 'CERTIFIED_ONLY'
    NOT_CERTIFIED_ONLY = 'NOT_CERTIFIED_ONLY'
    OFF = 'OFF'


class SearchPattern(IntEnum):
    """Enumeration constants used to specify searchType used to control BI
    Search. More details can be found in EnumDSSXMLSearchTypes in a browser."""

    CONTAINS_ANY_WORD = 0
    BEGIN_WITH = 1
    EXACTLY = 2
    BEGIN_WITH_PHRASE = 3
    CONTAINS = 4
    END_WITH = 5


class SearchDomain(IntEnum):
    """Enumeration constants used to specify the search domains. More details
    can be found in EnumDSSXMLSearchDomain in a browser."""

    LOCAL = 1
    PROJECT = 2
    REPOSITORY = 3
    CONFIGURATION = 4


class SearchResultsFormat(Enum):
    """Enumeration constants used to specify the format to return
    from search functions."""

    LIST = 'LIST'
    TREE = 'TREE'


class SearchScope(AutoUpperName):
    """Enumeration constants used to specify the scope of the search with regard
    to System Managed Objects."""

    ROOTED = auto()  # For compatibility
    NOT_MANAGED_ONLY = auto()
    MANAGED_ONLY = auto()
    ALL = auto()



# flake8: noqa
# this order of imports is important to avoid circular imports
from .applications import Application, list_applications
from mstrio.project_objects.bots import Bot, list_bots
from .agents import Agent, list_agents
from .content_cache import ContentCache
from .content_group import ContentGroup, list_content_groups
from .datasets import *
from .document import Document, list_documents, list_documents_across_projects
from .dashboard import (
    ChapterPage,
    Dashboard,
    DashboardChapter,
    PageSelector,
    PageVisualization,
    VisualizationSelector,
    list_dashboards,
    list_dashboards_across_projects,
)
from .incremental_refresh_report import (
    IncrementalRefreshReport,
    list_incremental_refresh_reports,
)
from .library import Library
from .report import Report, list_reports


from mstrio.project_objects.prompt import Prompt



from typing import TYPE_CHECKING

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error while creating the folder.")
def create_folder(
    connection: Connection,
    name: str,
    parent_id: str,
    description: str | None = None,
    project_id: str | None = None,
):
    """Create a folder.

    Note:
        When `project_id` is provided then folder will be created in this
        project. Otherwise it will be created in a project selected within
        `connection` object.

    Args:
        connection: Strategy One REST API connection object
        name (string): name of folder to create
        parent_id (string): id of folder in which new folder will be created
        description (string, optional): description of folder to create
        project_id (string, optional): id of project

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    body = {'name': name, 'description': description, 'parent': parent_id}
    return connection.post(
        endpoint='/api/folders', headers={'X-MSTR-ProjectID': project_id}, json=body
    )


@ErrorHandler(err_msg="Error while deleting folder with ID: {id}.")
def delete_folder(connection: Connection, id: str, project_id: str | None = None):
    """Delete complete folder.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project

    Returns:
        Complete Future object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        endpoint=f'/api/folders/{id}', headers={'X-MSTR-ProjectID': project_id}
    )


@ErrorHandler(err_msg="Error while listing folders.")
def list_folders(
    connection: Connection,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: str | None = None,
):
    """Get a list of folders.

    Args:
        connection: Strategy One REST API connection object
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/folders',
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def list_folders_async(
    future_session: 'FuturesSessionWithRenewal',
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get a list of folders asynchronously.

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            Strategy One REST Server asynchronously
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    endpoint = '/api/folders'
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(endpoint=endpoint, headers=headers, params=params)


@ErrorHandler(err_msg="Error while getting contents of a folder with ID: {id}.")
def get_folder_contents(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: str | None = None,
):
    """Get contents of a folder.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        endpoint=f'/api/folders/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def get_folder_contents_async(
    future_session: 'FuturesSessionWithRenewal',
    id: str,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get contents of a folder asynchronously.

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            Strategy One REST Server asynchronously
        id (string): ID of folder
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    project_id = (
        project_id if project_id is not None else future_session.connection.project_id
    )
    endpoint = f'/api/folders/{id}'
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(endpoint=endpoint, headers=headers, params=params)


@ErrorHandler(err_msg="Error while getting ID of a pre-defined folder.")
def get_predefined_folder_id(
    connection: Connection,
    folder_type: int,
    project_id: str | None = None,
):
    """Get ID of a pre-defined folder.

    Args:
        connection: Strategy One REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/folders/preDefined/{folder_type}/id',
        headers={'X-MSTR-ProjectID': project_id},
        params={'folderType': folder_type},
    )


@ErrorHandler(err_msg="Error while getting contents of a pre-defined folder.")
def get_predefined_folder_contents(
    connection: Connection,
    folder_type: int,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
    error_msg: str | None = None,
):
    """Get contents of a pre-defined folder.

    Args:
        connection: Strategy One REST API connection object
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/folders/preDefined/{folder_type}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'offset': offset, 'limit': limit},
    )


def get_predefined_folder_contents_async(
    future_session: 'FuturesSessionWithRenewal',
    folder_type: int,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Get contents of a pre-defined folder.

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            Strategy One REST Server asynchronously
        folder_type (int): predefined folder type, from `EnumDSSXMLFolderNames`
        project_id (string, optional): id of project
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 5000.

    Returns:
        Complete Future object.
    """
    endpoint = f'/api/folders/preDefined/{folder_type}'
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'offset': offset, 'limit': limit}
    return future_session.get(endpoint=endpoint, headers=headers, params=params)


@ErrorHandler(err_msg="Error while getting contents of My Personal Objects folder.")
def get_my_personal_objects_contents(
    connection: Connection, project_id: str | None = None
):
    """Get contents of My Personal Objects folder.

    Args:
        connection: Strategy One REST API connection object
        project_id (string, optional): id of project

    Returns:
         Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/folders/myPersonalObjects',
        headers={'X-MSTR-ProjectID': project_id},
    )



from mstrio.connection import get_connection
from mstrio.project_objects.incremental_refresh_report import ( 
    IncrementalRefreshReport,
    list_incremental_refresh_reports,
)
from mstrio.project_objects.datasets.helpers import (
    Template,
    AttributeTemplateUnit,
    MetricTemplateUnit,
    MetricElement,
)
from mstrio.project_objects.datasets.olap_cube import OlapCube


# Define variables which can be later used in a script
PROJECT_NAME = $project_name  # Name of project
FOLDER_ID = $folder_id  # Folder ID for listing and creating reports
FOLDER_PATH = $folder_path  # Folder path for listing reports
REPORT_NAME = $report_name  # Name of new report
CUBE_ID = $cube_id  # ID of target cube for new report

template_attribute_ids = $template_attribute_ids  # List of attribute IDs
template_metric_ids = $template_metric_ids  # List of metric IDs

template = Template(
    rows=[
        AttributeTemplateUnit(attribute_id) for attribute_id in template_attribute_ids
    ],
    columns=[
        MetricTemplateUnit([MetricElement(metric_id)])
        for metric_id in template_metric_ids
    ],
    page_by=[],  # page_by is optional if empty
)

conn = get_connection(workstationData, PROJECT_NAME)

# List all incremental refresh reports in a project
list_incremental_refresh_reports(conn, project_name=PROJECT_NAME)

# List all incremental refresh reports in a folder, by id
list_incremental_refresh_reports(conn, folder_id=FOLDER_ID)

# List all incremental refresh reports in a folder, by path
list_incremental_refresh_reports(conn, folder_path=FOLDER_PATH)

# Create an incremental refresh report given a cube
# This will use the cube's template.
new_report_from_cube = IncrementalRefreshReport.create_from_cube(
    conn,
    name=REPORT_NAME,
    destination_folder=FOLDER_ID,
    target_cube=OlapCube(conn, CUBE_ID),
    refresh_type=IncrementalRefreshReport.RefreshType.UPDATE_ONLY,
)

# Create an incremental refresh report given a cube
# and a custom template
new_report = IncrementalRefreshReport.create(
    conn,
    name=REPORT_NAME + 'Custom Template',
    destination_folder=FOLDER_ID,
    target_cube=OlapCube(conn, CUBE_ID),
    increment_type=IncrementalRefreshReport.IncrementType.REPORT,
    template=template,
    refresh_type=IncrementalRefreshReport.RefreshType.UPDATE_ONLY,
)

# Alter the incremental refresh report
new_template_attribute_ids = $new_template_attribute_ids  # List of attribute IDs
new_template_metric_ids = $new_template_metric_ids  # List of metric IDs

new_template = Template(
    rows=[
        AttributeTemplateUnit(attribute_id)
        for attribute_id in new_template_attribute_ids
    ],
    columns=[
        MetricTemplateUnit([MetricElement(metric_id)])
        for metric_id in new_template_metric_ids
    ],
)
new_report.alter(
    name=REPORT_NAME + 'altered',
    description='New description',
    refresh_type=IncrementalRefreshReport.RefreshType.UPDATE,
    template=new_template,
)

# Publish an incremental refresh report by ID
publish_id = $publish_id  # ID of report to publish
report_to_publish = IncrementalRefreshReport(conn, id=publish_id)
report_to_publish.execute()

# Delete an incremental refresh report by ID
delete_id = $delete_id  # ID of report to delete
report_to_delete = IncrementalRefreshReport(conn, id=delete_id)
report_to_delete.delete()



"""
    This example is for a dashboard user to modify data in a Rally workspace using
    a modern grid configured with transaction functions. The workflow will rely on this
    script to make actual modifications (update) in the defect.
"""

"""
    The properties available for modification in the defect are `Name`, `Description`,
    `Priority`, `Owner`, `Iteration`, `TargetDate`, `State` and `PlanEstimate`. Only
    the `Name` property is required. Property  `FormattedID` cannot be modified.

    To connect with Rally API we use the library 'pyral'. Python library to handle Rally
    connection is not installed on the environment by default so it is necessary
    to choose runtime which have such library installed in the settings of the script.
    Additionally it is necessary to specify correct 'Network Access' to allow script
    to communicate with some other IP then IP of the environment.
"""

from pyral import Rally, restapi
from datetime import datetime

# Helper method to return datetime object in correct format for Rally operations
def parse_date_time_value(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m/%d/%Y %H:%M:%S",
        "%m/%d/%Y",
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.strptime(dateValue, date_format)
            break
        except Exception:
            pass

    if parsed_datetime is None:
        raise Exception("Wrong Type of the datetime input value conversion!!!!")
    else:
        print(parsed_datetime)
        return parsed_datetime

try:
    # create connection to Rally. If you want to use other way then providing
    # api key please check documentation of this method. When providing secret keys like
    # `api_key` or any other sensitive data it is good to use variable of type `secret`.
    rally = Rally(
        server=$rally_server,
        apikey=$api_key,
        workspace=$workspace,
        project=$project,
    )

    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`.
    for index in range(len($formatted_id)):
        # UPDATE RALLY DEFECT
        tmp_formatted_id = $formatted_id[index]
        # first it is necessary to find object which will be modified
        defect = response_de = rally.get(
            "Defect",
            fetch=True,
            query=f"FormattedID = {tmp_formatted_id}",
            projectScopeDown=True,
        ).next()
        update_data = {"ObjectID": defect.ObjectID}
        if $defect_name and $defect_name[index]:
            update_data["Name"] = $defect_name[index]
        if $description:  # Value for description can be None
            update_data["Description"] = $description[index]
        # Valid 'Priority' values are [ Low, Normal, High Attention,
        # Resolve Immediately, None ]
        if $priority:
            update_data["Priority"] = $priority[index]
        # Value for owner can be None or existing owner username
        if $owner:
            owner = None
            # inside defect owner has to be stored as a reference to the actual user
            # value passed from dashboard is used to find this reference
            if $owner[index]:
                owner_username = $owner[index]
                owner = rally.get(
                    "User", fetch=True, query=f"UserName = {owner_username}"
                ).next().ref
            update_data["Owner"] = owner
        # Value for iteration can be None or existing iteration name
        if $iteration:
            iteration = None
            if $iteration[index]:
                tmp_iteration_name = $iteration[index]
                iteration = rally.get(
                    "Iteration", fetch=True, query=f'Name = "{tmp_iteration_name}"'
                ).next().ref
            update_data["Iteration"] = iteration
        if $target_date:
            target_date = (
                None
                if $target_date[index] is None
                else parse_date_time_value($target_date[index]).isoformat()
            )
            update_data["TargetDate"] = target_date
        # Valid 'State' values are: [ Submitted, Closed, Deferred, Open ]
        if $state and $state[index]:
            update_data["State"] = $state[index]
        if $plan_estimate:  # Value for 'Plan Estimate' can be None
            update_data["PlanEstimate"] = $plan_estimate[index]

        rally.update("Defect", update_data)
        
except restapi.RallyRESTAPIError as error:
    raise Exception(f"Error has occurred: {error}")
# this error is raised from `rally.get` when the object was not found and we are trying
# to iterate with `next` method
except StopIteration as error:
    raise Exception(f"Error while iterating has occurred: {error}")



"""This is the demo script to show how to manage applications.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.applications import Application, list_applications

# Define a variable which can be later used in a script
PROJECT_ID = $project_id  # Insert name of project here

conn = get_connection(workstationData, project_id=PROJECT_ID)


# Define variables which can be later used in a script
LIMIT = $limit
NAME_FILTER = $name_filter

# list applications with different conditions
# Note: No Applications exist in a default environment
list_of_all_apps = list_applications(connection=conn)
list_of_several_apps = list_applications(connection=conn, limit=LIMIT)
list_of_apps_name_filtered = list_applications(connection=conn, name=NAME_FILTER)
list_of_all_apps_as_dicts = list_applications(connection=conn, to_dictionary=True)

# Define variables which can be later used in a script
APP_NAME = $app_name
APP_DESCRIPTION = $app_description
APP_HOME_SCREEN_MODE = $app_home_screen_mode
APP_HOME_DOCUMENT_URL = $app_home_document_url
APP_HOME_DOCUMENT_TYPE = $app_home_document_type
APP_HOME_ICON = $app_home_icon
APP_HOME_TOOLBAR_MODE = $app_home_toolbar_mode
APP_HOME_LIBRARY_CONTENT_BUNDLE_ID = $app_home_library_content_bundle_id
APP_HOME_LIBRARY_SIDEBAR = $app_home_library_sidebar

# Create an Application
app = Application.create(
    connection=conn,
    name='Application Validation Scripts',
    description='This is a demo application created by the validation script.',
    home_screen=Application.HomeSettings(
        mode=APP_HOME_SCREEN_MODE,
        home_document=Application.HomeSettings.HomeDocument(
            url=APP_HOME_DOCUMENT_URL,
            home_document_type=APP_HOME_DOCUMENT_TYPE,
            icons=[APP_HOME_ICON],
            toolbar_mode=APP_HOME_TOOLBAR_MODE,
            toolbar_enabled=True,
        ),
        home_library=Application.HomeSettings.HomeLibrary(
            content_bundle_ids=[APP_HOME_LIBRARY_CONTENT_BUNDLE_ID],
            show_all_contents=False,
            icons=[APP_HOME_ICON],
            toolbar_mode=APP_HOME_TOOLBAR_MODE,
            sidebars=[APP_HOME_LIBRARY_SIDEBAR],
            toolbar_enabled=True,
        ),
    )
)

# Define variables which can be later used in a script
APP_ID = $app_id
APP_NEW_NAME = $app_new_name

# Get an Application by it's id or name
app = Application(connection=conn, id=APP_ID)
app = Application(connection=conn, name=APP_NAME)

# Alter an Application
app.alter(
    home_screen=app.home_screen,
    name=APP_NEW_NAME,
    general=app.general,
    email_settings=app.email_settings,
    ai_settings=app.ai_settings,
    auth_modes=app.auth_modes,
)

# Delete an Application without prompt
app.delete(force=True)



from logging import getLogger
from typing import TYPE_CHECKING

from requests import Response

from mstrio import config
from mstrio.api import palettes as palettes_api
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity
from mstrio.utils.format import Color
from mstrio.utils.helper import delete_none_values
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.server.project import Project

logger = getLogger(__name__)


@method_version_handler('11.3.0600')
def list_palettes(
    connection: Connection,
    to_dictionary: bool = False,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
) -> list['Palette'] | list[dict]:
    """List all palettes in the environment.

    Note:
        If no project related parameter is provided, configuration-level
        palettes will be listed.

    Args:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`
        to_dictionary: If True, returns a list of dictionaries, otherwise
            returns a list of Palette objects.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
    """
    return Palette._list_all(
        connection=connection,
        to_dictionary=to_dictionary,
        project=project,
        project_id=project_id,
        project_name=project_name,
    )


@class_version_handler('11.3.0600')
class Palette(Entity, CopyMixin, DeleteMixin):
    _OBJECT_TYPE = ObjectTypes.PALETTE
    _API_GETTERS = {
        (
            'id',
            'colors',
        ): palettes_api.get_palette,
        (
            'name',
            'description',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
            'hidden',
        ): objects_processors.get_info,
    }

    @staticmethod
    def _parse_colors(source: list[str], *args, **kwargs) -> list[Color]:
        return [Color(server_value=color) for color in source]

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'colors': _parse_colors,
    }
    _API_PATCH: dict = {
        (
            # the endpoint must take name, while not changing it
            # hence 'name' in both entries
            'name',
            'colors',
        ): (palettes_api.update_palette, 'put'),
        (
            'name',
            'description',
            'abbreviation',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Initialize a Palette object.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the palette.
            name (str, optional): Name of the palette.
            project (Project | str | None, optional): Project object or ID or
                name specifying the project. May be used instead of
                `project_id` or `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        if not id:
            if name:
                # not using find_object_with_name as api lacks search by name
                pals_with_name = [
                    c
                    for c in Palette._list_all(
                        connection=connection,
                        project_id=proj_id,
                        to_dictionary=True,
                    )
                    if c['name'] == name
                ]
                if not pals_with_name:
                    raise ValueError(f"Palette with name '{name}' not found.")
                if (num_of_pals := len(pals_with_name)) > 1:
                    raise ValueError(
                        f"Found {num_of_pals} palettes with name '{name}'. "
                        "Please provide 'id' argument."
                    )
                object_id = pals_with_name[0]['id']
            else:
                raise ValueError("Please provide either 'id' or 'name' argument.")
        else:
            object_id = id
        super().__init__(
            connection=connection,
            object_id=object_id,
            name=name,
            project_id=proj_id,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.colors = (
            Palette._parse_colors(cols)
            if (cols := kwargs.get('colors'))
            else default_value
        )

    @staticmethod
    def _color_to_server_value(color: Color | str) -> str:
        return color.server_value if isinstance(color, Color) else color

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        colors: list[Color] | list[str],
        abbreviation: str | None = None,
        description: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> 'Palette':
        """Create a new color palette.

        Note:
            If no project related parameter is provided, configuration-level
            palettes will be created.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            name (str): Name of the new palette.
            colors (list[Color] | list[str]): List of colors in the palette.
                In string format, colors should be in REST format, i.e. a
                single integer, e.g. "16737843" (16,737,843 = 0xff6633)
            abbreviation (str, optional): Abbreviation of the object name.
            description (str, optional): Description of the palette object.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            The newly created palette.
        """

        colors = [
            color.server_value if isinstance(color, Color) else color
            for color in colors
        ]
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )
        body = {
            'name': name,
            'colors': list(map(cls._color_to_server_value, colors)),
            'abbreviation': abbreviation,
            'description': description,
        }
        body = delete_none_values(source=body, recursion=True)
        res: Response = palettes_api.create_palette(
            connection=connection,
            body=body,
            project_id=proj_id,
        )
        new_id = res.json().get('id')
        new_pal = cls(
            connection=connection,
            id=new_id,
            project_id=proj_id,
        )
        if config.verbose:
            # duplicate name won't be accepted and won't create a changed name
            # so we can use the user-provided name
            logger.info(
                f"Successfully created Palette named: '{name}' with ID: '{new_id}'"
            )
        return new_pal

    def alter(
        self,
        name: str | None = None,
        colors: list[Color] | list[str] | None = None,
        abbreviation: str | None = None,
        description: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Alter the color palette's properties.
        Args:
            name (str, optional): New name of the palette.
            colors (list[Color] | list[str], optional): List of colors in the
                palette. In string format, colors should be in REST format,
                i.e. a single integer, e.g. "16737843" (16,737,843 = 0xff6633)
            abbreviation (str, optional): Abbreviation of the object name.
            description (str, optional): Description of the palette object.
            comments (str, optional): Long description of the palette.
            owner: (str | User, optional): Owner of the palette.
        """
        if isinstance(owner, User):
            owner = owner.id
        if name is None:
            name = self.name
        if colors:
            colors = [Palette._color_to_server_value(color) for color in colors]

        properties = {
            'name': name,
            'colors': colors,
            'abbreviation': abbreviation,
            'description': description,
            'comments': comments,
            'owner': owner,
        }
        properties = delete_none_values(properties, recursion=False)
        self._alter_properties(**properties)
        # update api does not return updated object data, only id;
        # need to fetch the object explicitly
        self.fetch()

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> list['Palette'] | list[dict]:
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
        )

        # not using fetch_objects as api lacks pagination
        res: Response = palettes_api.list_palettes(
            connection=connection,
            project_id=proj_id,
            fields='palettes',
        )
        objects = res.json().get('palettes', [])

        # listing api has 'paletteType': 0..1 instead of 'subtype': 17920..17921
        for obj in objects:
            obj['subtype'] = (
                obj.pop('paletteType') + ObjectSubTypes.SYSTEM_PALETTE.value
            )

        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]
